#!/usr/bin/env python
import hashlib
import os
import shutil
from collections import Counter
from datetime import datetime

import matplotlib.patches as mpatches
import matplotlib.pyplot as plt
import matplotlib.ticker as mtick
import numpy as np
import pandas as pd


def parse_aligns(args, input_dir):
    """
    Parses all fasta files in the user-given input directory
    Input: NONE
    Output: dictionary with genes a keys and SeqIO iterator as values
    """
    files = [os.path.join(input_dir, x) for x in os.listdir(input_dir) if x.endswith('.fas')]
    all_taxa_records = {}
    for file in files:
        gene = os.path.basename(file).split('.')[0]
        with open(file, 'r') as infile:
            for line in infile:
                line = line.strip()
                if line.startswith('>'):
                    taxon = line[1:]
                    if len(taxon.split('_')) == 4:
                        taxon = taxon.split('_')[0]

                    if taxon not in all_taxa_records.keys():
                        all_taxa_records[taxon] = [gene]
                    else:
                        all_taxa_records[taxon].append(gene)

    to_df = []
    for taxon in all_taxa_records:
        for gene in all_taxa_records[taxon]:
            to_df.append([taxon, gene])

    my_df = pd.DataFrame(to_df, columns=['UniqueID', 'Orthologs'])
    bin_mat = pd.crosstab(my_df.UniqueID, my_df.Orthologs)

    return bin_mat


def parse_metadata(metadata, input_meta=False):
    """
    Parses metadata.csv to get all org names in each group and subtax
    Input: NONE
    Output: (1) dictionary with groups/subtax as keys and sets of orgs in those groups/subtax as values
            (2) list of all orgs in metadata
    """
    taxa_dict = {}
    with open(metadata, 'r') as infile:
        for line in infile:
            line = line.strip()
            if "Unique ID" not in line:
                if input_meta:
                    _, _, org, group, subtax, _, long_name, _, _ = line.split('\t')
                else:
                    org, long_name, group, subtax, _, _ = line.split('\t')
                taxa_dict[org] = [group, subtax, long_name]

    return taxa_dict


def completeness(args, input_dir, genes=True):
    """
    Computes completeness of genes based on either all taxa_comp_df or a set of taxa_comp_df provided by user
    """

    matrix = parse_aligns(args, input_dir)
    taxa_count, gene_count = matrix.shape
    gene_comp = matrix.sum().divide(other=taxa_count)
    taxa_comp = matrix.sum(axis=1).divide(other=gene_count)

    if genes:
        return matrix
    else:
        return taxa_comp, len(gene_comp)


def make_plot(s, plot_name, y_count, genes=True):
    """
    Creates a plot of gene completeness. Plot is a bar chart with genes sorted from highest to lowest completeness on
        the x-axis and percent complete on the y-axis. Completeness is calculated as percent of taxa_comp_df the gene is present
        in.
    Input: Pandas DataFrame, plot name
    Output: PDF of plot
    """

    if genes:
        calculated_s = 'Ortholog'
        calculated_p = 'Orthologs'
        calc_by = 'Taxa'
    else:
        calculated_s = 'Taxon'
        calculated_p = 'Taxa'
        calc_by = 'Orthologs'

    s = s.sort_values(ascending=False)
    # Dictionary with completeness as key and color as threshold. Color will be used to color bars that are above
    #   threshold.
    colors_threshold = {0.0: 'coral', 0.1: 'black', 0.2: 'lime', 0.3: 'maroon', 0.4: 'khaki',
                        0.5: 'purple', 0.6: 'orange', 0.7: 'olive', 0.8: 'grey', 0.9: 'navy'}
    colors_threshold_inv = {v: k for k, v in colors_threshold.items()}

    # Assigns color to genes depending on the completeness. Based off dict above
    s_colors = [''] * len(s.index)
    for i, row in enumerate(s):
        for j, threshold in enumerate(colors_threshold.keys()):
            if row >= threshold:  # Assigns color based on completeness value
                s_colors[i] = colors_threshold[threshold]
    df = s.to_frame()
    df['colors'] = s_colors  # Adds color list as a column in this DataFrame
    df_colors = [x for x in df['colors']]  # Converts list to tuple because df.plot requires tuple not list
    threshold_counts = Counter(df['colors'])

    # initializing plot
    fig = plt.figure()
    ax = fig.add_subplot(111)
    s.plot(width=0.85,
           color=df_colors,
           figsize=(35, 10),
           legend=False,
           fontsize=8,
           ax=ax,
           kind='bar')

    ax.set_title(f'{calculated_s} Completeness by {calc_by}', fontsize=30)

    # Legend
    legend_data = []
    for i, x in enumerate(sorted(colors_threshold.keys(), reverse=True)):
        if i == len(threshold_counts):
            break
        if i == 0:
            label_str = f'Completeness ≥ {round(x * 100)}% ({threshold_counts[colors_threshold[x]]} {calculated_p})'
        else:
            label_str = (f'{round(x * 100)}% ≤ Completeness < {round((x + 0.1) * 100)}%'
                         f' ({threshold_counts[colors_threshold[x]]} {calculated_p})')
        legend_data.append(mpatches.Patch(color=colors_threshold[x],
                                          label=label_str))

    ax.legend(handles=legend_data, facecolor='white', framealpha=1)

    # Y-axis settings and formatting
    ax.set_ylabel(f'Percent Complete ({y_count} {calc_by})', labelpad=15, fontsize=20)
    ax.yaxis.set_major_formatter(mtick.PercentFormatter(1.0))
    ax.set_ylim(0, 1)
    ax.yaxis.set_ticks(np.arange(0, 1.1, 0.1))
    ax.tick_params(axis='y', which='both', labelleft=True, labelright=True)

    # X-axis settingsand formattings
    ax.set_xlabel(f'{calculated_p} ({len(df.index)})', labelpad=15, fontsize=20)

    # Horizontal Lines
    for x in threshold_counts.keys():  # Iterate through colors with counts being the values
        plt.axhline(y=colors_threshold_inv[x], xmin=0, xmax=1, color=x, zorder=0)
    plt.axhline(y=1, xmin=0, xmax=1, color='black', zorder=0)

    # Top X-Axis
    gene_num = len(df.index) - 1
    ax2 = ax.twiny()
    ax2.set_xlim(ax.get_xlim())
    ax2.set_xticks(range(-1, gene_num, 5))
    ax2.set_xticklabels(range(0, gene_num, 5))
    ax2.xaxis.set_minor_locator(mtick.MultipleLocator(1))

    # Graphic rendering
    fig.set_tight_layout(True)
    fig.savefig(f'{plot_name}.pdf')
    fig.show()


def get_md5(filename):
    # Open,close, read file and calculate MD5 on its contents
    with open(filename, 'rb') as file_to_check:
        # read contents of the file
        data = file_to_check.read()
        # pipe contents of the file through
        md5_returned = hashlib.md5(data).hexdigest()

    return md5_returned


def first_backup(dfo, date_time):
    os.mkdir(f'{dfo}/backups')
    os.mkdir(f'{dfo}/backups/{date_time}')
    shutil.copytree(f'{dfo}/orthologs', f'{dfo}/backups/{date_time}/orthologs')
    shutil.copytree(f'{dfo}/paralogs', f'{dfo}/backups/{date_time}/paralogs')
    shutil.copytree(f'{dfo}/proteomes', f'{dfo}/backups/{date_time}/proteomes')
    shutil.copy(f'{dfo}/metadata.tsv', f'{dfo}/backups/{date_time}/metadata.tsv')
    shutil.copy(f'{dfo}/tree_colors.tsv', f'{dfo}/backups/{date_time}/tree_colors.tsv')


def backup(dfo):
    now = datetime.now().strftime('%d-%b-%Y_%H-%M-%S')

    if os.path.isdir(f'{dfo}/backups'):
        backups = os.listdir(f'{dfo}/backups')
        backups.sort(key=lambda date: datetime.strptime(date, '%d-%b-%Y_%H-%M-%S'), reverse=True)
        latest_backup = backups[0]

        back_files = {}
        db_files = {}
        os.mkdir(f'{dfo}/backups/{now}')
        os.mkdir(f'{dfo}/backups/{now}/orthologs')
        os.mkdir(f'{dfo}/backups/{now}/paralogs')
        os.mkdir(f'{dfo}/backups/{now}/proteomes')

        for root, dirs, files in os.walk(f'{dfo}/backups'):
            for file in files:
                back_files[file] = (os.path.join(root, file))

        for folder in ['orthologs', 'paralogs', 'proteomes']:
            for root, dirs, files in os.walk(f'{dfo}/{folder}'):
                for file in files:
                    db_files[file] = (os.path.join(root, file))

        db_files['metadata.tsv'] = f'{dfo}/metadata.tsv'
        db_files['tree_colors.tsv'] = f'{dfo}/tree_colors.tsv'

        for k, v in db_files.items():
            if k == 'metadata.tsv' or k == 'tree_colors.tsv':
                dest = k
            else:
                dest = '/'.join(v.split('/')[-2:])

            if k in back_files.keys() and get_md5(db_files[k]) == get_md5(back_files[k]):
                os.symlink(f'{dfo}/backups/{latest_backup}/{dest}', f'{dfo}/backups/{now}/{dest}')
            else:
                shutil.copy(v, f'{dfo}/backups/{now}/{dest}')

    else:
        first_backup(dfo, now)



#!/usr/bin/env python
import configparser
import os
import subprocess
import textwrap
from pathlib import Path
from phylofisher import help_formatter

SNAKEFILE_PATH = f'{os.path.dirname(os.path.realpath(__file__))}/sgt_constructor.smk'

def bash(cmd):
    """
    Function to run bash commands in a shell
    :param cmd:
    :return:
    """
    subprocess.run(cmd, shell=True, executable='/bin/bash')  # ,


def get_genes():
    return [file.split('.')[0] for file in os.listdir(args.input)]


def get_outfiles():
    if args.no_trees:
        outfiles = [f'{args.output}/trimal/{gene}.final' for gene in get_genes()]
    else:
        outfiles = [f'{args.output}-local.tar.gz']

    return ' '.join(outfiles)
    

def make_config():
    config_frags = [
        f'out_dir={args.output}',
        f'in_dir={args.input}',
        f'genes={",".join(get_genes())}',
        f'metadata={args.metadata}',
        f'tree_colors={args.color_conf}',
        f'input_metadata={args.input_metadata}',
        f'trees_only={args.trees_only}',
        f'no_trees={args.no_trees}'
    ]

    return ' '.join(config_frags)


def run_snakemake():
    smk_frags = [
        f'snakemake',
        f'-s {SNAKEFILE_PATH}',
        f'--config {make_config()}',
        f'--cores {args.threads}',
        f'--rerun-incomplete',
        f'--keep-going',
        f'--nolock'
    ]

    if args.dry_run:
        smk_frags.append('-n')
        
    smk_cmd = ' '.join(smk_frags)
    smk_cmd += ' ' + get_outfiles()
    bash(smk_cmd)


if __name__ == '__main__':
    description = 'Aligns, trims, and builds single gene trees from unaligned gene files.'
    usage = 'sgt_constructor.py -i path/to/input/ [OPTIONS]'
    parser, optional, required = help_formatter.initialize_argparse(name='sgt_constructor.py',
                                                                    desc=description,
                                                                    usage=usage)

    # Optional Arguments
    optional.add_argument('-t', '--threads', metavar='N', type=int, default=1,
                          help=textwrap.dedent("""\
                          Desired number of threads to be utilized.
                          Default: 1
                          """))
    optional.add_argument('--no_trees', action='store_true',
                          help=textwrap.dedent("""\
                          Do NOT build single gene trees.
                          Length filtration and trimmming only.
                          """))
    optional.add_argument('--trees_only', action='store_true',
                          help=textwrap.dedent("""\
                          Only build single gene trees.
                          No length filtration and trimming.
                          """))
    optional.add_argument('--dry_run', action='store_true',
                          help=textwrap.dedent("""\
                          Builds DAG and show jobs that would be performed.
                          """))
    optional.add_argument('-if', '--in_format', metavar='<format>', type=str, default='fasta',
                          help=textwrap.dedent("""\
                          Format of the input files.
                          Options: fasta, phylip (names truncated at 10 characters), 
                          phylip-relaxed (names are not truncated), or nexus.
                          Default: fasta
                          """))

    args = help_formatter.get_args(parser, optional, required, pre_suf=False, inp_dir=True)

    config = configparser.ConfigParser()
    config.read('config.ini')
    dfo = str(Path(config['PATHS']['database_folder']).resolve())
    args.metadata = str(os.path.join(dfo, 'metadata.tsv'))
    args.color_conf = str(os.path.abspath(config['PATHS']['color_conf']))
    args.input_metadata = str(os.path.abspath(config['PATHS']['input_file']))
    args.input = str(os.path.abspath(args.input[:-1])) if args.input.endswith('/') else str(os.path.abspath(args.input))
    args.output = str(os.path.abspath(args.output[:-1])) if args.output.endswith('/') else str(os.path.abspath(args.output))

    out_dict = {'fasta': 'fas',
                'phylip': 'phy',
                'phylip-relaxed': 'phy',
                'nexus': 'nex'}

    run_snakemake()


#!/usr/bin/env python
import configparser
import os
import shutil
import sys
import tarfile
import textwrap
from datetime import date
from pathlib import Path

import pandas as pd
from Bio import SeqIO
from Bio.Seq import Seq
from Bio.SeqRecord import SeqRecord

from phylofisher import help_formatter, tools


def parse_collapse_tsv():
    with open(args.input, 'r') as infile:
        collapse_dict = dict()
        for line in infile:
            collapse_dict[line.strip().split('\t')[0]] = line.strip().split('\t')[1:]
    return collapse_dict


def check_metadata():
    """
    Checks to make sure taxa provided in to_collapse.csv are actually in the meta data
    """
    with open(args.input, 'r') as infile:
        to_collapse = set()
        for taxon in [line.strip().split(',')[2:] for line in infile]:
            to_collapse.update(taxon)

    taxon_dict = dict.fromkeys(to_collapse, False)

    with open(metadata, 'r') as infile:
        for line in infile:
            line = line.strip()

            short_name = line.split('\t')[0]
            if short_name in to_collapse:
                taxon_dict[short_name] = True

    for taxon in taxon_dict.keys():
        if taxon_dict[taxon] is False:
            print(f'{taxon} is not in the dataset.')

    if False in taxon_dict.values():
        sys.exit('Taxa not collapsed. Please fix errors above.')


def add_to_meta(abbrev, input_info):
    """"Transfer input metadata for a given organism
    to dataset metadata.
    input:  short name of organism
    return: None
    """

    new_row = {'Unique ID': abbrev,
               'Long Name': input_info[abbrev][0],
               'Higher Taxonomy': input_info[abbrev][1],
               'Lower Taxonomy': input_info[abbrev][2],
               'Source': 'taxon_collapser.py',
               'Data Type': 'Collapsed'
               }

    df = pd.read_csv(metadata, delimiter='\t')
    df.index = df['Unique ID']

    df = df.append(new_row, ignore_index=True)
    df = df.sort_values(by=['Higher Taxonomy', 'Lower Taxonomy', 'Unique ID'])
    df.to_csv(str(Path(dfo, 'metadata.tsv')), sep='\t', index=False)


def collapse():
    collapse_dict = parse_collapse_tsv()

    # Collapse Ortholgogs
    ortho_dir = f'{dfo}/orthologs'
    gene_files = [os.path.join(ortho_dir, gene) for gene in os.listdir(ortho_dir) if gene.endswith('.fas')]
    for gene_file in gene_files:
        record_dict = {k: SeqRecord(Seq(''), id='', name='', description='') for k in collapse_dict.keys()}
        records = []
        with open(gene_file, 'r') as infile:
            for record in SeqIO.parse(infile, 'fasta'):
                for key in collapse_dict.keys():
                    if record.description in collapse_dict[key]:
                        if len(record.seq) > len(record_dict[key].seq):
                            record_dict[key].id = key
                            record_dict[key].seq = record.seq

                records.append(record)

        for record in record_dict.values():
            if str(record.seq) != '':
                records.append(record)

        with open(gene_file, 'w') as outfile:
            SeqIO.write(records, outfile, 'fasta')

        record_dict.clear()

    # Collapse Paralogs
    para_dir = f'{dfo}/paralogs'
    gene_files = [os.path.join(para_dir, gene) for gene in os.listdir(para_dir) if gene.endswith('.fas')]
    for gene_file in gene_files:
        records = []
        with open(gene_file, 'r') as infile:
            for record in SeqIO.parse(infile, 'fasta'):
                for key in collapse_dict.keys():
                    if record.description.split('..')[0] in collapse_dict[key]:
                        records.append(SeqRecord(record.seq, id=f'{key}..{record.description.split("..")[1]}', name='', description=''))
                
                records.append(record)

        with open(gene_file, 'w') as outfile:
            SeqIO.write(records, outfile, 'fasta')

        record_dict.clear()

    for key in collapse_dict.keys():
        add_to_meta(key, collapse_dict)


if __name__ == '__main__':
    description = 'Collapses dataset entries into a single taxon'
    parser, optional, required = help_formatter.initialize_argparse(name='taxon_collapser.py',
                                                                    desc=description,
                                                                    usage='taxon_collapser.py [OPTIONS] -i <taxa>')

    required.add_argument('-i', '--input', type=str, metavar='to_collapse.tsv', default=None,
                          help=textwrap.dedent("""\
                          A .tsv containing a Unique ID, higher and lower taxonomic designations, 
                           and the Unique IDs of the taxa to collapse, for each chimera one per line"""))
    args = help_formatter.get_args(parser, optional, required,
                                   out_dir=False, pre_suf=False, inp_dir=False)

    # Config Parser
    config = configparser.ConfigParser()
    config.read('config.ini')
    dfo = str(Path(config['PATHS']['database_folder']).resolve())
    metadata = str(os.path.join(dfo, 'metadata.tsv'))

    tools.backup(dfo)
    check_metadata()
    collapse()



#!/usr/bin/env python
import configparser
import os, sys
import textwrap
from glob import glob
from pathlib import Path

import pandas as pd
from Bio import SeqIO

from phylofisher import help_formatter

pd.options.display.float_format = '{:,.0f}'.format


class Metadata:
    """
    Class which store and combine information from metadata.tsv and 
    orthologs and paralogs from database.
    """

    def __init__(self, metadata, ortholog_folder, paralog_folder):
        self.metadata = pd.read_csv(metadata, sep='\t')

        self.ortholog_folder = ortholog_folder
        self.paralog_folder = paralog_folder
        self.metadata['Orthologs'] = self.metadata['Unique ID'].map(self.parse_orthologs())
        self.metadata['Paralogs'] = self.metadata['Unique ID'].map(self.parse_paralogs())

        self.org_meta = self.metadata[['Unique ID', 'Long Name', 'Higher Taxonomy', 'Lower Taxonomy',
                                       'Data Type', 'Orthologs', 'Paralogs', 'Source']]
        self.metadata = self.metadata[['Unique ID', 'Long Name', 'Higher Taxonomy', 'Lower Taxonomy',
                                       'Data Type', 'Orthologs', 'Paralogs']]


    def parse_orthologs(self):
        """
        Count orthologs for all organisms.
        return: dictionary[org] = count
        """
        counted_orthologs = {}
        for file in glob(f'{self.ortholog_folder}/*.fas'):
            for record in SeqIO.parse(file, 'fasta'):
                if record.description not in counted_orthologs:
                    counted_orthologs[record.description] = 1
                else:
                    counted_orthologs[record.description] += 1
        return counted_orthologs

    def parse_paralogs(self):
        """
        Count paralogs for all organisms.
        return: dictionary[org] = count
        """
        counted_paralogs = {}
        for file in glob(f'{self.paralog_folder}/*.fas'):
            for record in SeqIO.parse(file, 'fasta'):
                name = record.description.split('.')[0]
                if name not in counted_paralogs:
                    counted_paralogs[name] = 1
                else:
                    counted_paralogs[name] += 1
        return counted_paralogs

    def higher_taxonomy(self):
        """
        Return table with metadata grouped according to a higher taxonomy.
        """
        return self.metadata.groupby(['Higher Taxonomy']
                                     ).size().reset_index().rename(columns={0: 'Organisms'})

    def lower_taxonomy(self):
        """
        Return table with metadata grouped according to a lower taxonomy.
        """
        return self.metadata.groupby(['Higher Taxonomy', 'Lower Taxonomy']
                                     ).size().reset_index().rename(columns={0: 'Organisms'})

    def get_higher(self, term):
        """
        Return table with orthologs, paralogs for organisms from a given Higher Taxonomy.
        """
        try:
            ret = self.metadata[self.metadata['Higher Taxonomy'] == term].sort_values(['Lower Taxonomy', 'Unique ID']).reset_index(drop=True)
        except KeyError:
            sys.exit(f'{term} is not a valid Higher Taxonomy')
        return ret

    def get_lower(self, term):
        """
        Return table with orthologs, paralogs for organisms from a given Lower Taxonomy.
        """
        try:
            ret = self.metadata[self.metadata['Lower Taxonomy'] == term].sort_values('Unique ID').reset_index(drop=True)
        except KeyError:
            sys.exit(f'{term} is not a valid Lower Taxonomy')
        return ret 

    def get_org(self, term):
        """
        Return ifnormation about a given organism using its short name.
        """
        org_df = self.org_meta.copy()
        org_df = org_df.set_index('Unique ID')
        try:
            df = pd.DataFrame([org_df.loc[term, :]]).transpose()
        except KeyError:
            sys.exit(f'{term} is not a valid UniqueID')
        return df


if __name__ == '__main__':
    description = 'Explore database'
    parser, optional, required = help_formatter.initialize_argparse(name='explore_database.py',
                                                                    desc=description,
                                                                    usage='explore_database.py [OPTIONS]')

    # Optional Arguments
    optional.add_argument('-d', '--database', type=str,  # metvar='<path/to/database>',
                          help=textwrap.dedent("""\
                          Path to the database folder if config.ini is not used.
                              """))
    optional.add_argument('-t', '--higher_taxonomy', action='store_true',
                          help=textwrap.dedent("""\
                          Show table with metadata grouped according to the higher taxonomy.
                              """))

    optional.add_argument('-l', '--lower_taxonomy', action='store_true',
                          help=textwrap.dedent("""\
                          Show table with metadata grouped according to the lower taxonomy.
                              """))

    optional.add_argument('-r', '--get_higher', default=None,
                          help=textwrap.dedent("""\
                          Return table with orthologs, paralogs for organisms from a given Higher Taxonomy.
                              """))

    optional.add_argument('-w', '--get_lower', default=None,
                          help=textwrap.dedent("""\
                          Return table with orthologs, paralogs for organisms from a given Lower Taxonomy.
                              """))

    optional.add_argument('-o', '--get_org', default=None,
                          help=textwrap.dedent("""\
                          Return information about a given organism using its short name.
                              """))

    args = help_formatter.get_args(parser, optional, required, pre_suf=False, out_dir=False, inp_dir=False)

    config = configparser.ConfigParser()
    config.read('config.ini')

    if args.database:
        dfo = str(Path(args.database).resolve())
    else:
        dfo = str(Path(config['PATHS']['database_folder']).resolve())

    ortholog_folder = str(Path(dfo, f'orthologs'))
    paralog_folder = str(Path(dfo, f'paralogs'))
    metadata = str(os.path.join(dfo, 'metadata.tsv'))
    metadata_handle = Metadata(metadata, ortholog_folder, paralog_folder)

    if args.higher_taxonomy:
        print(metadata_handle.higher_taxonomy().to_string(index=False))

    elif args.lower_taxonomy:
        print(metadata_handle.lower_taxonomy().to_string(index=False))

    elif args.get_higher:
        print(metadata_handle.get_higher(args.get_higher).to_string(index=False))

    elif args.get_lower:
        print(metadata_handle.get_lower(args.get_lower).to_string(index=False))

    elif args.get_org:
        print(metadata_handle.get_org(args.get_org).to_string())



import argparse
import textwrap
from datetime import date


version = '1.0.14'
today = date.today()


class CustomHelpFormatter(argparse.HelpFormatter):
    """
    This class changes the way the help output is displayed
    """

    def _format_action_invocation(self, action):
        # This removes metvar after short option
        if not action.option_strings or action.nargs == 0:
            return super()._format_action_invocation(action)
        default = self._get_default_metavar_for_optional(action)
        args_string = self._format_args(action, default)
        return ', '.join(action.option_strings) + ' ' + args_string

    def _split_lines(self, text, width):
        # This adds 3 spaces before lines that wrap
        lines = text.splitlines()
        for i in range(0, len(lines)):
            if i >= 1:
                lines[i] = (3 * ' ') + lines[i]
        return lines


class MyHelpFormatter(CustomHelpFormatter, argparse.RawTextHelpFormatter):
    pass


def add_global_arguments(parser, optional, required, in_help, out_help, inp, pre_suf, out_dir):
    """
    Function to add argparse arguments used in all scripts

    input: argparse group class instance "optional"
           argparse group class instance "required"
           in_help = help message output for input
           out_help = help message output for output
    """

    # Required
    if inp is True:
        required.add_argument('-i', '--input', required=True, type=str, metavar='<in_dir>',
                              help=textwrap.dedent(f"""{in_help}"""))

    # Optional
    if out_dir is True:
        today_date = today.strftime("%b.%d.%Y")
        out_dir = f'{format(parser.prog).split(".")[0]}_out_{today_date}'
        optional.add_argument('-o', '--output', default=out_dir, type=str, metavar='<out_dir>',
                              help=textwrap.dedent(f"""{out_help}
                                      """))
    if pre_suf is True:
        optional.add_argument('-p', '--prefix', metavar='<prefix>', type=str, default='',
                              help=textwrap.dedent("""\
                              Prefix of input files
                              Default: NONE
                              Example: path/to/input/prefix*"""))
        optional.add_argument('-s', '--suffix', metavar='<suffix>', type=str, default='',
                              help=textwrap.dedent("""\
                              Suffix of input files.
                              Default: NONE
                              Example: path/to/input/*suffix
                              """))
    optional.add_argument('-h', '--help', action='help', default=argparse.SUPPRESS,
                          help=textwrap.dedent("""\
                                      Show this help message and exit.
                                      """))


def initialize_argparse(name: object, desc: object, usage: object) -> object:
    """
    This function initialized argparse
    """
    formatter = lambda prog: MyHelpFormatter(prog, max_help_position=100)

    # noinspection PyTypeChecker
    parser = argparse.ArgumentParser(prog=name,
                                     description=desc,
                                     usage=usage,
                                     formatter_class=formatter,
                                     add_help=False,
                                     epilog=textwrap.dedent(f"""\
                                         additional information:
                                            Version: {version}
                                            GitHub: https://github.com/TheBrownLab/PhyloFisher
                                            Cite: doi:https://10.1371/journal.pbio.3001365
                                            """))

    optional = parser._action_groups.pop()
    required = parser.add_argument_group('required arguments')

    return parser, optional, required


def get_args(parser, optional, required, pre_suf=True, inp_dir=True, out_dir=True,
             in_help='Path to input directory'):

    out_help = ('Path to user-defined output directory\n'
                f'Default: ./{format(parser.prog).split(".")[0]}_out_<M.D.Y>')

    add_global_arguments(parser, optional, required,
                         in_help=in_help, out_help=out_help, inp=inp_dir,
                         pre_suf=pre_suf, out_dir=out_dir)

    parser._action_groups.append(optional)

    return parser.parse_args()



#!/usr/bin/env python
import os
import shutil
import textwrap
from statistics import mean

from Bio import SeqIO
from ete3 import Tree

from phylofisher import help_formatter


class Leaves:

    def __init__(self, tree):
        self.tree = tree
        self.leaf_names = self.tree.get_leaf_names()
        self.leaves = self.tree.get_leaves()

    def get_locations(self):
        locs = {}
        for name in self.leaf_names:
            locs[name] = self.tree & name
        return locs

    def get_mean_distance(self, leaf):
        locations = self.get_locations()
        distances = []
        for name in self.leaf_names:
            if name != leaf.name:
                distances.append(leaf.get_distance(locations[name]))
        sorted_dist = sorted(distances, reverse=True)
        return mean(sorted_dist[:   10])

    def org_speed(self):
        mean_distances = []
        for leaf in self.leaves:
            mean_distances.append((self.get_mean_distance(leaf), leaf.name))
        sorted_dist = sorted(mean_distances, reverse=True)
        return [i[1] for i in sorted_dist]


class Matrix:

    def __init__(self, matrix, format, ranked_orgs):
        self.matrix = matrix
        self.format = format
        self.ranked_orgs = ranked_orgs

    def fast_evol_taxa(self, step_size):
        steps = []
        for i in range(0, len(self.ranked_orgs), step_size):
            steps.append(self.ranked_orgs[0:i + step_size])
        return steps

    def generate_subset(self, output_folder, iterations, step_size):
        steps = self.fast_evol_taxa(step_size)
        os.mkdir(output_folder)
        for i in range(0, iterations):
            exclude = steps[i]

            # Original Code. Does NOT re-align and trim
            #
            # output_name = os.path.join(output_folder, f'step{i}')
            # with open(output_name, 'w') as res:
            #     for record in SeqIO.parse(self.matrix, self.format):
            #         if record.name not in exclude:
            #             res.write(f'>{record.name}\n{record.seq}\n')

            seq_files = [file for file in os.listdir(args.ortholog_files) if file.endswith('.fas')]

            os.mkdir(f'{output_folder}/tmp')

            for file in seq_files:
                with open(f'{args.ortholog_files}/{file}', 'r') as infile, open(f'{output_folder}/tmp/{file}', 'w') as outfile:
                    records = []
                    for record in SeqIO.parse(infile, 'fasta'):
                        if record.name not in exclude:
                            records.append(record)

                    SeqIO.write(records, outfile, 'fasta')

            os.system(f'matrix_constructor.py '
                      f'-i {output_folder}/tmp '
                      f'-o {output_folder}/step_{i} '
                      f'-of {args.out_format.lower()} '
                      f'-t {args.threads}')

            shutil.rmtree(f'{output_folder}/tmp')


if __name__ == '__main__':
    description = 'Removes the fastest evolving taxa, based on branch length.'
    parser, optional, required = help_formatter.initialize_argparse(name='fast_taxa_remover.py',
                                                                    desc=description,
                                                                    usage='fast_taxa_remover.py [OPTIONS] '
                                                                          '-tr <tree> '
                                                                          '-m <matrix> '
                                                                          '-i <num_of_iterations>')
    # Required Arguments
    required.add_argument('-m', '--matrix', required=True, type=str, metavar='matrix',
                          help=textwrap.dedent("""\
                          Path to input matrix.
                          """))
    required.add_argument('-tr', '--tree', required=True, type=str, metavar='tree',
                          help=textwrap.dedent("""\
                          Path to input tree.
                          """))
    required.add_argument('-i', '--iterations', metavar='N', type=int, required=True,
                          help=textwrap.dedent("""\
                          Number of iterations
                          """))
    required.add_argument('-or', '--ortholog_files', metavar='N', type=str, required=True,
                          help=textwrap.dedent("""\
                          Path to directory containing single gene files. This will be th path to 
                          prep_final_dataset_<M.D.Y> if used within the main PhyloFisher workflow.
                          """))

    # Optional Arguments
    optional.add_argument('-in_format', metavar='<format>', type=str, default='fasta',
                          help=textwrap.dedent("""\
                          Input matrix format if not FASTA.
                          Options: fasta, phylip (names truncated at 10 characters), 
                          phylip-relaxed (names are not truncated), or nexus.
                          Default: fasta
                          """))
    optional.add_argument('-out_format', metavar='<format>', type=str, default='fasta',
                          help=textwrap.dedent("""\
                          Desired output format.
                          Options: fasta, phylip (names truncated at 10 characters),
                          phylip-relaxed (names are not truncated), or nexus.
                          Default: fasta
                          """))
    optional.add_argument('-s', '--step_size', metavar='N', type=int, default=1,
                          help=textwrap.dedent("""\
                          Number taxa removed per iteration.
                          Default: 1
                          """))
    optional.add_argument('-t', '--threads', metavar='N', type=int, default=1,
                          help=textwrap.dedent("""\
                          Desired number of threads to be utilized.
                          Default: 1
                          """))

    args = help_formatter.get_args(parser, optional, required, pre_suf=False, inp_dir=False)

    tree = Tree(args.tree)
    x = Leaves(tree)
    m = Matrix(args.matrix, args.in_format, x.org_speed())
    m.generate_subset(args.output, args.iterations, args.step_size)



#!/usr/bin/env python
import csv
import os
import random
import shutil
import string
import subprocess
import sys
import textwrap
from collections import defaultdict, Counter
from glob import glob

import pandas as pd
from Bio import SeqIO

from phylofisher import help_formatter


def csv_to_tsv():
    csvs = [csv for csv in os.listdir('.') if csv.endswith('csv')]
    for csv in csvs:
        with open(csv, 'r') as infile, open(f'{csv.split(".")[0]}.tsv', 'w')as outfile:
            for line in infile:
                line = line.strip().replace(',', '\t')
                outfile.write(f'{line}\n')
        os.remove(csv)

    tsvs = [tsv for tsv in os.listdir('.') if tsv.endswith('tsv')]
    for tsv in tsvs:
        with open(tsv, 'r') as infile, open('tmp', 'w') as outfile:
            for line in infile:
                line = line.strip().replace(',', '\t')
                outfile.write(f'{line}\n')
        shutil.move('tmp', tsv)


def rename_taxa():
    """"""


    # Read in to_rename.tsv
    rename_dict = dict()
    with open(args.rename, 'r') as infile:
        infile.readline()
        for line in infile:
            line = line.strip()
            line_list = line.split('\t')
            rename_dict[line_list[0]] = line_list[1:]

    # Rename in orthologs/ and paralogs/
    files = glob('orthologs/*.fas') + glob('paralogs/*.fas')
    for file in files:
        with open(file, 'r') as infile, open('tmp', 'w') as tmp_file:
            for line in infile:
                line = line.strip()
                for key, value in rename_dict.items():
                    if key in line:
                        line = line.replace(key, value[0])
                tmp_file.write(f'{line}\n')
        shutil.move('tmp', file)

    df = pd.read_csv('metadata.tsv', sep='\t')
    for key, value in rename_dict.items():
        df.loc[df['Unique ID'] == key, 'Long Name'] = value[1]
        df.loc[df['Unique ID'] == key, 'Unique ID'] = value[0]
    
    df.to_csv('metadata.tsv', sep='\t', index=False)


def get_ortho_taxa():
    """
    Returns unique set of taxa in orthologs dir
    """
    unique_orgs_orthos = set()
    files = glob('orthologs/*.fas')
    for file in files:
        with open(file, 'r') as infile:
            records = SeqIO.parse(infile, 'fasta')
            for record in records:
                unique_orgs_orthos.add(record.description)

    return unique_orgs_orthos


def check_orthologs():
    """
    Check that IDs in orthologs are all unique in the file.
    """
    files = glob('orthologs/*.fas')
    for file in files:
        ids = set()
        with open(file, 'r') as infile:
            records = SeqIO.parse(infile, 'fasta')
            for record in records:
                if record.name in ids:
                    print(f"ERROR: {record.name} is not a unique ID in {file}")
                    sys.exit()
                ids.add(record.name)


def get_meta_taxa():
    """
    Returns unique set of taxa in metadata
    """
    unique_orgs_meta = set()
    with open('metadata.tsv', 'r') as csvfile:
        reader = csv.reader(csvfile, delimiter='\t')
        for row in reader:
            id_ = row[0]
            if id_ == 'Unique ID':
                continue
            
            #check that ID is unique
            if id_ in unique_orgs_meta:
                print("ERROR: ", id_, "is not a unique ID")
                sys.exit()

            #check illegal characters
            for ch in ["_", '@', '..', '*', ' ']:
                if ch in id_:
                    print(f"ERROR: illegal character {ch} in {id_}")
                    sys.exit()

            unique_orgs_meta.add(id_)
    return unique_orgs_meta


def check_taxa():
    """
    Checks to see if all taxa in metadata is in ortholog dir and vice-versa. If there are differences the script exits
    here and prints the discrepancies.

    :return: None
    """
    unique_orgs_orthos = get_ortho_taxa()
    unique_orgs_meta = get_meta_taxa()

    # Exits if there are differences and prints those differences
    if len(unique_orgs_meta - unique_orgs_orthos) > 0 or len(unique_orgs_orthos - unique_orgs_meta) > 0:
        if len(unique_orgs_meta - unique_orgs_orthos) > 0:
            print(f'Taxa in metadata but not in orthologs dir:')
            for org in (unique_orgs_meta - unique_orgs_orthos):
                print(org)
        if len(unique_orgs_orthos - unique_orgs_meta) > 0:
            print(f'Taxa in orthlog dir but not in metadata:')
            for org in (unique_orgs_orthos - unique_orgs_meta):
                print(org)
        sys.exit()


def id_generator(size=5, chars=string.digits):
    """
    Generate random number with 5 digits.
    :param size:
    :param chars:
    :return:
    """
    return ''.join(random.choice(chars) for _ in range(size))


def paralog_name(abbrev, keys):
    """"Prepare paralog name (short name + 5 random digits).
    Recursive function.
    example: Homosap..12345
    input: short name of an organism, names of already existing paralogs
    for a given organism
    return: unique paralog name"""
    id_ = id_generator()
    pname = f'{abbrev}..p{id_}'
    if pname not in keys:
        return pname
    else:
        paralog_name(abbrev, keys)


def prepare_diamond_input():
    """
    Prepares seqs for diamond in a way that it connects
    information about gene before sequence name.
    Example: Allomacr from ADK2.fas will be named
    ADK2@Allomacr at the for_diamond.fasta file.

    :return: None
    """
    files = glob('orthologs/*.fas')
    with open("for_diamond.fasta", 'w') as res:
        for file in files:
            gene = file.split('/')[-1].split('.')[0]
            for record in SeqIO.parse(file, 'fasta'):
                res.write(f'>{gene}@{record.description}\n{record.seq}\n')


def diamond():
    """
    Uses for_diamond.fasta to diamond orthomcl.diamonddb

    :return: None
    """
    with open('for_diamond.fasta', 'r') as infile, open('new.fas', 'w') as outfile:
        for line in infile:
            line = line.strip()
            if line.startswith(">") is False:
                line = line.replace('-', '')
            outfile.write(f'{line}\n')

    shutil.move('new.fas', 'for_diamond.fasta')

    db = 'orthomcl/orthomcl.diamonddb.dmnd'
    out = 'diamond.res'
    cmd = (
        f'diamond blastp -e 1e-10 -q for_diamond.fasta --more-sensitive '
        f'--db {db} -o {out} -p {args.threads} --outfmt 6 qseqid stitle evalue')
    subprocess.run(cmd, shell=True)


def parse_diamond_output():
    gene_ogs = defaultdict(list)
    for line in open('diamond.res'):
        sline = line.split("\t")
        full_name = sline[0]  # ADK2@Allomacr
        gene, _ = full_name.split('@')  # gene: ADK2
        og = sline[1].split('|')[2].strip()  # parse og name OG5_128398
        gene_ogs[gene].append(og)  # gene_ogs['ADK2'].append(og)
    return gene_ogs


def genes_in_orthodb():
    """
    Check if gene has an og in orthomcl. If not -> report this gene.
    """
    gene_in_gene_og = set()
    for line in open("orthomcl/gene_og"):
        gene_in_gene_og.add(line.split()[0])

    files = glob('orthologs/*.fas')
    genes = [file.split('/')[-1].split('.')[0] for file in files]
    rerun = False
    for gene in genes:
        if gene not in gene_in_gene_og:
            print(f"{gene} has not hit in orthomcl and thus can not be used in databases."
            "\nPlease delete this gene in orthologs and re-run build_database.py.")
            rerun = True
    if rerun == True:
        sys.exit()


def get_og_file(threshold):
    prepare_diamond_input()  # prepares orthologs for diamondF
    diamond()  # starts diamond
    gene_filtered_ogs = defaultdict(list)
    gene_ogs = parse_diamond_output()
    for gene, ogs in gene_ogs.items():
        total = len(ogs)
        counted = dict(Counter(ogs))
        for og, count in counted.items():
            if count / total >= threshold:
                gene_filtered_ogs[gene].append(og)
    with open('orthomcl/gene_og', 'w') as res:
        for gene, ogs in gene_filtered_ogs.items():
            res.write(f'{gene}\t{",".join(ogs)}\n')
    os.remove('diamond.res')
    os.remove('for_diamond.fasta')


def concat_gene_files():
    files = glob('*.fas')

    with open('datasetdb.fasta', 'w') as outfile:
        for file in files:
            with open(file, 'r') as infile:
                for line in infile:
                    line = line.strip()
                    if line.startswith('>'):
                        line = f'>{file.split(".")[0]}'
                    outfile.write(f'{line}\n')


def datasetdb():
    # TODO fix me
    os.mkdir('datasetdb')
    os.chdir('orthologs')
    concat_gene_files()
    shutil.move('datasetdb.fasta', '../datasetdb')
    os.chdir('../datasetdb')
    dmd_db = 'diamond makedb --in datasetdb.fasta -d datasetdb'
    subprocess.run(dmd_db, shell=True)
    # os.remove('datasetdb.fasta')
    os.chdir('..')


def make_profiles(threads):
    """
    Prepares hmm profiles from all fasta files with orthologs
    from orthologs folder.

    :param threads: number of threads
    :return: None
    """

    os.mkdir('profiles')
    os.chdir('orthologs')

    # prepares alignments for all fasta files with orthologs
    aln = f'for i in $(ls *.fas); do mafft --auto' \
          f' --thread {threads} --reorder $i > $(basename $i .fas).aln; done'
    subprocess.run(aln, shell=True)

    # prepares hmm profiles from alignmets made in prev. step
    hmm = 'for i in $(ls *.aln); do hmmbuild $(basename $i .aln).hmm $i; done'
    subprocess.run(hmm, shell=True)
    subprocess.run('mv *.hmm ../profiles', shell=True)
    subprocess.run('rm *.aln', shell=True)  # delete alignments
    os.chdir('..')


def main(args, threads, no_og_file, threshold):
    """
    :param args:
    :param threads: number of threads
    :param make_og_file: Boolean
    :param threshold: float 0-1
    :return: None
    """

    # checks if paralogs folder exists and creates
    # it if not
    if os.path.isdir('paralogs') is False:
        os.mkdir('paralogs')
    if os.path.isdir('datasetdb') is True:
        shutil.rmtree('datasetdb')
    if os.path.isdir('profiles') is True:
        shutil.rmtree('profiles')

    if args.rename:
        rename_taxa()

    datasetdb()  # creates datasetdb

    make_profiles(threads)

    if not no_og_file:
        get_og_file(threshold)

    genes_in_orthodb()

if __name__ == "__main__":
    description = 'Script for database construction and taxonomic updates. Must be run within path/to/database/'
    parser, optional, required = help_formatter.initialize_argparse(name='build_database.py',
                                                                    desc=description,
                                                                    usage='build_database.py [OPTIONS]')

    optional.add_argument('-t', '--threads', type=int, default=1,
                          help=textwrap.dedent("""\
                          Number of threads.
                          Default:1
                          """))

    optional.add_argument('-n', '--no_og_file', action='store_true',
                          help=textwrap.dedent("""\
                          Do not make Gene OG file
                          """))
    optional.add_argument('-o', '--og_threshold', type=float, default=0.1, metavar='0.X',
                          help=textwrap.dedent("""\
                          0.X proportion of sequences (0-1) of sequences that must hit an 
                          OrthoMCL orthogroup for the group to be assigned.
                          Default: 0.1 (10%%)
                          """))
    optional.add_argument('--rename', type=str, metavar='<to_rename.tsv>',
                          help=textwrap.dedent("""\
                          Rename taxa in the database.
                          Takes a TSV file containing the Old Unique ID, New Unique ID, and New Long Name.
                          """))

    args = help_formatter.get_args(parser, optional, required, pre_suf=False, inp_dir=False, out_dir=False)

    csv_to_tsv()
    check_orthologs()
    check_taxa()
    main(args, args.threads, args.no_og_file, args.og_threshold)



#!/usr/bin/env python
import os, re
import shutil
import subprocess
import textwrap

import numpy as np
import pandas as pd

from phylofisher import help_formatter


def run_raxml_rtc():
    """
    :return:
    """
    if os.path.isdir(f'{args.output}/RAxML'):
        shutil.rmtree(f'{args.output}/RAxML')
    os.mkdir(f'{args.output}/RAxML')
    os.chdir(f'{args.output}/RAxML')
    genes = [file.split('.')[1] for file in os.listdir(args.input) if 'RAxML_bipartitions.' in file]
    genes = set(genes)

    for gene in genes:
        cmd = (f'raxmlHPC-PTHREADS-AVX2 '
               f'-t {args.input}/RAxML_bipartitions.{gene}{args.suffix} '
               f'-z {args.input}/RAxML_bootstrap.{gene}{args.suffix} '
               f'-n IC.{gene}{args.suffix} '
               f'-f i '
               f'-m PROTCATLGF '
               f'-T 2 '
               f'-s NA')

        subprocess.run(cmd, shell=True, executable='/bin/bash')

    os.chdir(cwd)


def get_rtc():
    """

    :return:
    """
    genes = [file.split('.')[1] for file in os.listdir(args.input) if 'RAxML_bipartitions.' in file]
    rtc_dict = {}
    for gene in genes:
        with open(f'{args.output}/RAxML/RAxML_info.IC.{gene}{args.suffix}', 'r') as infile:
            for line in infile:
                line = line.strip()
                if 'Tree certainty for this tree:' in line:
                    rtc = float(line.split(': ')[1])
                    rtc_dict[gene] = rtc

    return rtc_dict


def get_genes():
    """

    :return:
    """
    s = pd.Series(rtc_values)
    s = s.sort_values(ascending=False)
    # Get top 25%, 50%, and 75%
    s25 = s[s > np.percentile(s, 75)]
    s50 = s[s > np.percentile(s, 50)]
    s75 = s[s > np.percentile(s, 25)]
    # Get genes in each of the following pandas series
    genes_25 = s25.index
    genes_50 = s50.index
    genes_75 = s75.index

    return {'25': genes_25, '50': genes_50, '75': genes_75}


def build_matrices():
    gene_subsets = get_genes()
    gene_files = [os.path.join(args.input, file) for file in os.listdir(args.input)
                  if file.endswith(out_dict[args.in_format])]

    for key, subset in gene_subsets.items():
        if os.path.isdir(f'{args.output}/rtc{key}'):
            shutil.rmtree(f'{args.output}/rtc{key}')
        os.mkdir(f'{args.output}/rtc{key}')
    
        for gene_file in gene_files:
            for gene in subset:
                if gene in gene_file:
                    dst = f'{args.output}/rtc{key}/{os.path.basename(gene_file)}'
                    shutil.copy(gene_file, dst)

        cmd = (f'matrix_constructor.py '
               f'-i {args.output}/rtc{key} '
               f'-o {args.output}/matrix_constructor{key} '
               f'-if {args.in_format.lower()} '
               f'-of {args.out_format.lower()} '
               f'-c')
        subprocess.run(cmd, shell=True, executable='/bin/bash')


if __name__ == '__main__':
    description = ('Bins orthologs based on relative tree certainty score (RTC) calculated in RAxML \n'
                   'and constructs supermatrices from the bins.')
    parser, optional, required = help_formatter.initialize_argparse(name='rtc_binner.py',
                                                                    desc=description,
                                                                    usage='rtc_binner.py '
                                                                          '[OPTIONS] -i /path/to/input/')

    # Optional Arguments
    optional.add_argument('-if', '--in_format', metavar='<format>', type=str, default='fasta',
                          help=textwrap.dedent("""\
                          Format of the input alignments.
                          Options: fasta, nexus, phylip (names truncated at 10 characters),
                          or phylip-relaxed (names are not truncated)
                          Default: fasta
                          """))

    optional.add_argument('-of', '--out_format', metavar='<format>', type=str, default='fasta',
                          help=textwrap.dedent("""\
                          Desired format of the RTC binned alignments.
                          Options: fasta, nexus, phylip (names truncated at 10 characters), 
                          or phylip-relaxed (names are not truncated)
                          Default: fasta
                          """))

    optional.add_argument('-s', '--suffix', metavar='<suffix>', type=str, default='NONE',
                          help=textwrap.dedent("""\
                          Suffix of RAxML output files
                          Example: RAxML_bipartitions.GENE.<SUFFIX>, RAxML_bootstrap.GENE.<SUFFIX>
                          Default: NONE
                          """))

    in_help = ('Path to directory containing single gene trees built from only orthologs, \n'
               'corresponding bootstrap value files, and corresponding alignments.')
    args = help_formatter.get_args(parser, optional, required, in_help=in_help, pre_suf=False)

    args.output = os.path.abspath(args.output)
    args.input = os.path.abspath(args.input)

    out_dict = {'fasta': 'fas',
                'phylip': 'phy',
                'phylip-relaxed': 'phy',
                'nexus': 'nex'}

    if args.suffix and args.suffix[0] != '.':
        args.suffix = f'.{args.suffix}'

    if os.path.isdir(args.output):
        shutil.rmtree(args.output)
    os.mkdir(args.output)

    cwd = os.getcwd()

    run_raxml_rtc()
    rtc_values = get_rtc()
    build_matrices()



#!/usr/bin/env python
import configparser
import os
import sys
import textwrap
from collections import Counter
from pathlib import Path

import matplotlib.pyplot as plt
import pandas as pd
from ete3 import Tree

from phylofisher import help_formatter

plt.rcParams["figure.figsize"] = (10, 5)


def get_taxa_set(trees):
    """
    """


def bipartitions(tree):
    """

    :param tree:
    :return:
    """
    bipar = set()
    all_ = frozenset(tree.get_leaf_names())
    for node in tree.traverse('preorder'):
        if node.is_leaf() is False:
            taxa = frozenset(node.get_leaf_names())
            rest = frozenset(taxa.symmetric_difference(all_))
            if len(taxa) > 1:
                bipar.add(taxa)
            if len(rest) > 1:
                bipar.add(rest)
    return bipar


def support(trees):
    """

    :param trees:
    :return:
    """
    bootstrap = []
    n_trees = 0
    all_taxa = set()
    for line in open(trees):
        line = line.strip()
        leaves = list(Tree(line).get_leaf_names())
        all_taxa.update(leaves)

    for line in open(trees):
        line = line.strip('')
        tree = Tree(line)
        n_trees += 1
        bootstrap = bootstrap + list(bipartitions(tree))
    norm_bootstrap = {}
    counted = dict(Counter(bootstrap))
    for key, value in counted.items():
        norm_bootstrap[key] = value / n_trees
    
    return norm_bootstrap, all_taxa


def get_support(group, supp_dict):
    """

    :param group:
    :param supp_dict:
    :return:
    """
    query = frozenset(group)
    if query in supp_dict:
        return supp_dict[query]
    else:
        return 0


def get_taxa_in_group(groups):
    """

    :return:
    """
    df = pd.read_csv(metadata, delimiter='\t')
    taxa = []
    for group in groups:
        if group in list(df['Higher Taxonomy']):
            taxa = taxa + list(df.loc[df['Higher Taxonomy'] == group, 'Unique ID'])
        elif group in list(df['Lower Taxonomy']):
            taxa = taxa + list(df.loc[df['Lower Taxonomy'] == group, 'Unique ID'])
        else:
            sys.exit(f'{group} is not in the database\'s metadata')

        if args.chimeras:
            for chimera in chim_dict.keys():
                if chim_dict[chimera][0] == group or chim_dict[chimera][1] == group:
                    taxa.append(chimera)

    return taxa


def parse_groups(input_file):
    """

    :param input_file:
    :return:
    """
    query_dict = {}
    for line in open(input_file):
        line = line.strip()
        if ':' in line:
            group = line.split(':')[0]
            orgs = line.split(':')[1]
            query_dict[group] = [org.strip() for org in orgs.split(',')]
        else:
            group = line
            groups = [group for group in group.split('+')]
            query_dict[group] = get_taxa_in_group(groups)
    return query_dict.items()


def file_to_series(file):
    """

    :param file:
    :return:
    """
    print(file)
    sup_dict, all_taxa = support(file)
    group_sup = {}
    for group, orgs in queries:
        orgs = set(orgs) - ((set(orgs) - all_taxa))
        group_sup[group] = get_support(orgs, sup_dict)
    s = pd.Series(group_sup)
    return s


def parse_bss():
    """

    :return:
    """
    columns = []
    n = 0
    for line in open(args.bs_files):
        line = line.strip()
        column = file_to_series(line.strip())
        column.name = os.path.basename(line)
        # column.name = f'Step {n}'
        columns.append(column)
        n += 1
        
    return columns


def main():
    """

    :return:
    """
    columns = parse_bss()
    df = pd.DataFrame(columns)

    fig = plt.figure()
    ax = fig.add_subplot(111)

    if args.bar_plot:
        df = df.transpose()
        df.plot.bar(figsize=(20, 10),
                    legend=True,
                    fontsize=12,
                    ax=ax,
                    cmap="Set2")
        ax.set_xlabel('Groups', size=16)

    else:
        df.plot(figsize=(35, 10),
                legend=True,
                fontsize=8,
                ax=ax,
                cmap="Set2")
        ax.set_xticks(range(len(list(df.index))))

    ax.set_xticklabels(list(df.index), rotation=45, ha='right')
    ax.set_ylabel('Bootstrap Support', size=12)
    ax.legend(loc=0, prop={'size': 12})
    ax.margins(x=0.005)
    plt.tight_layout()
    fig.savefig("bipartition_examiner.pdf")
    df.to_csv("bipartition_examiner.tsv", sep='\t')
    return df


if __name__ == "__main__":
    description = 'Calculates the observed occurrences of clades of interest in bootstrap trees.'
    parser, optional, required = help_formatter.initialize_argparse(name='bipartition_examiner.py',
                                                                    desc=description,
                                                                    usage='bipartition_examiner.py '
                                                                          '[OPTIONS] -i /path/to/input/')

    # Required Arguments
    required.add_argument('-b', '--bs_files', required=True, type=str, metavar='',
                          help=textwrap.dedent("""\
                          Path bootstrap files.
                          """))
    required.add_argument('-g', '--groups', type=str, required=True, metavar='',
                          help=textwrap.dedent("""\
                          groups
                          """))

    # Optional Arguments
    optional.add_argument('-on', '--output_name', type=str, metavar='out', default='out',
                          help=textwrap.dedent("""\
                          """))
    optional.add_argument('--database', type=str, metavar='path/to/db',
                          help=textwrap.dedent("""\
                          Path to database if not using config.ini
                          """))
    optional.add_argument('--no_db', action='store_true',
                          help=textwrap.dedent("""\
                          
                          """))
    optional.add_argument('--chimeras', type=str, metavar='chimera_info.tsv',
                          help=textwrap.dedent("""\
                          Path to TSV file containing taxonomic designations for chimeras.
                          """))
    optional.add_argument('--bar_plot', action='store_true',
                          help=textwrap.dedent("""\
                          Plot categorical data as a barplot.
                          Default: Plot series data as a line graph.
                          """))

    args = help_formatter.get_args(parser, optional, required, pre_suf=False, inp_dir=False)

    if not args.no_db:
        if args.database:
            dfo = os.path.abspath(args.database)
        else:
            config = configparser.ConfigParser()
            config.read('config.ini')
            dfo = str(Path(config['PATHS']['database_folder']).resolve())

        metadata = str(os.path.join(dfo, 'metadata.tsv'))

    if args.chimeras:
        chim_dict = {}
        with open(args.chimeras, 'r') as infile:
            for line in infile:
                line = line.strip()
                split_line = line.split('\t')
                chim_dict[split_line[0]] = split_line[1:3]

    queries = parse_groups(args.groups)
    main()



#!/usr/bin/env python
import os
import random
import string
import subprocess
import sys
import textwrap

import pandas as pd
from Bio import SeqIO
from Bio.Seq import Seq
from Bio.SeqRecord import SeqRecord
from ete3 import Tree

from phylofisher import help_formatter


def bash(cmd):
    """
    Runs bash commands as a subproccess
    @param cmd:
    @return: NONE
    """
    subprocess.run(cmd, executable='/bin/bash', shell=True)


def id_generator(size=10, chars=string.digits):
    return ''.join(random.choice(chars) for _ in range(size))


def unique_name(keys):
    id_ = id_generator()
    if id_ not in keys:
        return id_
    else:
        unique_name(keys)


def fake_phylip(matrix):
    length = None
    pseudonames = {}
    pseudonames_rev = {}
    records = []

    for record in SeqIO.parse(matrix, args.in_format):
        uname = unique_name(pseudonames)
        pseudonames[record.name] = uname
        pseudonames_rev[uname] = record.name
        seq = str(record.seq)
        records.append(SeqRecord(Seq(seq),
                                 id=uname,
                                 name='',
                                 description=''))

    SeqIO.write(records, 'renamed.phy', 'phylip')

    return pseudonames, pseudonames_rev


def fake_tree(treefile, pseudonames):
    with open('renamed.tre', 'w') as res:
        original = open(treefile).readline()
        for key, value in pseudonames.items():
            original = original.replace(key, value)
        res.write(original)


def get_taxa():
    """
    Parses input supermatrix to retrieve taxa
    @return: list of taxa and a dictionary with names as keys and as pandas series of sites as the keys
    """
    taxa_list = []
    mat_dict = {}
    with open('renamed.phy', 'r') as infile:
        for record in SeqIO.parse(infile, 'phylip'):
            taxa_list.append(record.description)
            mat_dict[record.name] = pd.Series(list(record.seq))
    return taxa_list, mat_dict


def get_branch_lens():
    """
    Parse input tree and retrieve branch lengths
    @return:
    """
    tree = Tree('renamed.tre')
    dist_mat = []
    for i, taxon1 in enumerate(taxa):
        dist_mat.append([])
        node1 = tree & taxon1
        for taxon2 in taxa:
            node2 = tree & taxon2
            dist = tree.get_distance(node1, node2)
            dist_mat[i].append(dist)
    length = int((len(taxa) - 1) / 2)
    taxa_dict = {}
    for i, taxon in enumerate(taxa):
        bran_lens = sorted(dist_mat[i], reverse=True)[0:length]
        taxa_dict[taxon] = sum(bran_lens) / length
    series = pd.Series(taxa_dict)
    size = int(len(taxa) / 3)

    return series, size


def prune_tree(speed):
    """

    @param speed:
    @return:
    """
    # Prune tree into SLOW tree
    if speed == 'fast':
        mylist = list(series.nlargest(size).index)
    else:
        mylist = list(series.nsmallest(size).index)
    tree = Tree('renamed.tre')
    nodes = [tree & x for x in mylist]
    tree.prune(nodes, preserve_branch_length=True)
    tree.write(outfile=f'{speed}.tre')

    return mylist


def trim_matrix():
    """

    @return:
    """
    with open('renamed.phy', 'r') as infile, open('fast.phy', 'w') as fast_fas, open('slow.phy', 'w') as slow_fas:
        fast_recs = []
        slow_recs = []
        for record in SeqIO.parse(infile, 'phylip'):
            if record.description in slow_taxa:
                slow_recs.append(record)
            elif record.description in fast_taxa:
                fast_recs.append(record)

        SeqIO.write(slow_recs, slow_fas, 'phylip')
        SeqIO.write(fast_recs, fast_fas, 'phylip')


def get_site_rates():
    """

    @return:
    """
    for speed in ['slow', 'fast']:
        with open(f'{speed}.dist_est.ctl', 'w') as outfile:
            ctl = (f'treefile = {speed}.tre * treefile\n'
                   f'seqfile = {speed}.phy * sequence data\n'
                   'nchar = 20             * amino acid data\n'
                   'model =  3             * empirical + F\n'
                   'aaRatefile = lg.dat * JTT substitution model\n'

                   'nrate = 101            * number of rates\n'
                   'ub = 10.0              * upper bound for rates\n')
            outfile.write(ctl)

    # Calculate site rates for FAST taxa
    bash(f'dist_est fast.dist_est.ctl')
    os.rename('rate_est.dat', 'fast.rate_est.dat')
    os.rename('DE.dat', 'fast.DE.dat')

    # Calculate site rates for SLOW taxa
    bash(f'dist_est slow.dist_est.ctl')
    os.rename('rate_est.dat', 'slow.rate_est.dat')
    os.rename('DE.dat', 'slow.DE.dat')


def parse_rates(speed):
    """

    @param speed:
    @return:
    """
    if speed == 'fast':
        rate_est = 'fast.rate_est.dat'
    else:
        rate_est = 'slow.rate_est.dat'

    rates = []
    for line in open(rate_est):
        _, myrate, _, _ = line.split()
        rates.append(myrate)

    return rates


def get_site_ratio():
    """

    @return:
    """
    rates_ratio = [float(x) / float(y) for x, y in zip(fast_rates, slow_rates)]
    positions = []
    for i, rate in enumerate(rates_ratio):
        positions.append((i, float(rate)))
    sorted_positions = sorted(positions, key=lambda x: x[1], reverse=True)
    result = [i[0] for i in sorted_positions]

    return result


def site_removal():
    """

    @return:
    """
    os.mkdir(f'steps_{args.step_size}')
    os.chdir(f'steps_{args.step_size}')

    out_dict = {'fasta': 'fas',
                'phylip': 'phy',
                'phylip-relaxed': 'phy',
                'nexus': 'nex'}

    for i, step in enumerate(range(args.step_size, len(sorted_sites), args.step_size)):
        with open(f'step{i}.{out_dict[args.in_format]}', 'w') as outfile:
            records = []
            for name, seq in matrix_dict.items():
                seq = "".join(seq[sorted_sites[step:]].values)
                records.append(SeqRecord(Seq(seq),
                                         id=name_dict_rev[name],
                                         name='',
                                         description=''))

            # Writes to output matrix in user specified output
            if args.out_format.lower() in out_dict:
                SeqIO.write(records, outfile, args.out_format.lower())
            else:
                sys.exit('Invalid Output Format')


if __name__ == '__main__':
    description = ('Removes the most heterotachous sites within a phylogenomic supermatrix in a stepwise fashion,'
                   ' leading to a user defined set of new matrices with these sites removed.')
    usage = 'heterotachy.py -t path/to/tree -m path/to/matrix [OPTIONS]'
    parser, optional, required = help_formatter.initialize_argparse(name='heterotachy.py',
                                                                    desc=description,
                                                                    usage=usage)

    required.add_argument('-tr', '--tree', type=str, metavar='', required=True,
                          help=textwrap.dedent("""\
                          Path to tree
                          """))
    required.add_argument('-m', '--matrix', type=str, metavar='', required=True,
                          help=textwrap.dedent("""\
                          Path to supermatrix
                          """))
    # Optional Arguments

    optional.add_argument('-s', '--step_size', type=int, default=3000, metavar='N',
                          help=textwrap.dedent("""\
                          Size of removal step (i.e., 1000 sites removed) to exhaustion
                          Default: 3000
                          """))
    optional.add_argument('-if', '--in_format', metavar='<format>', type=str, default='fasta',
                          help=textwrap.dedent("""\
                          Input format of matrix
                          Options: fasta, nexus, phylip (names truncated at 10 characters), 
                          or phylip-relaxed (names are not truncated)
                          Default: fasta
                          """))
    optional.add_argument('-f', '--out_format', metavar='<format>', type=str, default='fasta',
                          help=textwrap.dedent("""\
                          Desired format of the output steps.
                          Options: fasta, nexus, phylip (names truncated at 10 characters), 
                          or phylip-relaxed (names are not truncated)
                          Default: fasta
                          """))

    args = help_formatter.get_args(parser, optional, required, pre_suf=False, inp_dir=False)

    cwd = os.getcwd()

    if not os.path.exists(args.output):
        os.mkdir(args.output)
    os.chdir(args.output)

    name_dict, name_dict_rev = fake_phylip(f'{cwd}/{args.matrix}')
    fake_tree(f'{cwd}/{args.tree}', name_dict)

    # Parse input alignment
    taxa, matrix_dict = get_taxa()

    # Get Branch Lengths
    series, size = get_branch_lens()

    # Prune tree
    slow_taxa = prune_tree('slow')
    fast_taxa = prune_tree('fast')

    # Slow and Fast fastas
    trim_matrix()

    get_site_rates()
    fast_rates = parse_rates('fast')
    slow_rates = parse_rates('slow')

    sorted_sites = get_site_ratio()
    site_removal()



#!/usr/bin/env python
import random
import string
import subprocess
import textwrap

from Bio import SeqIO
from Bio.SeqRecord import SeqRecord

from phylofisher import help_formatter


def id_generator(size=10, chars=string.ascii_uppercase):
    return ''.join(random.choice(chars) for _ in range(size))


def unique_name(keys):
    id_ = id_generator()
    if id_ not in keys:
        return id_
    else:
        unique_name(keys)


def fake_phylip(matrix):
    seqs = 0
    length = None
    pseudonames = {}
    pseudonames_rev = {}
    records = []
    for record in SeqIO.parse(matrix, args.in_format):
        seqs += 1
        if not length:
            length = len(record.seq)
        uname = unique_name(pseudonames)
        pseudonames[record.name] = uname
        pseudonames_rev[uname] = record.name
        records.append(SeqRecord(record.seq,
                                 id=uname,
                                 name='',
                                 description=''))
    SeqIO.write(records, 'TEMP.phy', 'phylip-sequential')
    return pseudonames, pseudonames_rev


def fake_tree(treefile, pseudonames):
    with open('TEMP.tre', 'w') as res, open('key.tsv', 'w') as key_tsv:
        original = open(treefile).readline()
        key_tsv.write('Original Name\tID\n')
        for key, value in pseudonames.items():
            original = original.replace(key, value)
            key_tsv.write(f'{key}\t{value}\n')
        res.write(original)


def run_mammal():
    cmd = f'mammal -s TEMP.phy -t TEMP.tre -c {args.rate_classes} -l'
    subprocess.run(cmd, executable='/bin/bash', shell=True)


if __name__ == "__main__":
    description = 'Prepares supermatrix and tree for MAMMaL analysis'
    parser, optional, required = help_formatter.initialize_argparse(name='mammal_modeler.py',
                                                                    desc=description,
                                                                    usage='mammal_modeler.py '
                                                                          '[OPTIONS] -t <tree_file> -s <matrix>')

    # Required Arguments
    required.add_argument('-m', '--matrix', required=True, type=str, metavar='<matrix>',
                          help=textwrap.dedent("""\
                              Path to supermatrix file.
                              """))
    required.add_argument('-tr', '--tree', type=str, metavar='<tree>',
                          help=textwrap.dedent("""\
                              Path to tree.
                              """))
    # Optional Arguments
    optional.add_argument('-c', '--rate_classes', metavar='<N>', type=int, default=60,
                          help=textwrap.dedent("""\
                              The number of frequency classes in the mixture model.
                              Options: 10, 20, 30, 40, 50, or 60
                              Default: 60
                            """))
    optional.add_argument('-if', '--in_format', metavar='<format>', type=str, default='fasta',
                          help=textwrap.dedent("""\
                              Input format of matrix
                              Options: fasta, nexus, phylip (names truncated at 10 characters), 
                              or phylip-relaxed (names are not truncated)
                              Default: fasta
                              """))

    args = help_formatter.get_args(parser, optional, required, pre_suf=False, inp_dir=False)

    pseudo_, pseudo_rev_ = fake_phylip(args.matrix)
    fake_tree(args.tree, pseudo_)
    run_mammal()



#!/usr/bin/env python
import configparser
import os
import textwrap
from pathlib import Path

from phylofisher import help_formatter, tools


def in_out_completeness(df):
    """

    :param df:
    :return:
    """
    # Gets all taxa unique IDS
    all_taxa_set = set()
    with open(f'{dfo}/metadata.tsv', 'r') as infile:
        for line in infile:
            line = line.strip().split('\t')[0]
            all_taxa_set.add(line)

    # Gets out group Unique IDS
    out_group_set = set()
    with open(args.out_group, 'r') as infile:
        for line in infile:
            line = line.strip()
            out_group_set.add(line)

    # Determine in group from all taxa - out group
    in_group_set = all_taxa_set - out_group_set

    # Calculates completeness for each gene based on in group presence
    in_df = df[df.index.isin(list(in_group_set))]
    in_df = in_df.sum().divide(other=len(in_df))

    # Calculates completeness for each gene based on out group presence
    out_df = df[df.index.isin(list(out_group_set))]
    out_df = out_df.sum().divide(other=len(out_df))

    return out_df.to_dict(), in_df.to_dict()


def make_subset_tsv():
    """

    :return: df
    """
    df = gene_comp.to_frame()
    df = df.rename(columns={0: 'Completeness'})
    df = df.sort_values(by=['Completeness'], ascending=False)
    df = df.round({'Completeness': 3})

    if args.out_group:
        out_dict, in_dict = in_out_completeness(matrix)
        df['In-Group Completeness'] = df.index.map(in_dict)
        df['Out-Group Completeness'] = df.index.map(out_dict)
        df = df.round({'In-Group Completeness': 3})
        df = df.round({'Out-Group Completeness': 3})

    to_keep = ['yes' for k in df.index]
    df['Include in Subset'] = to_keep

    if args.gene_number:
        num_to_keep = args.gene_number
        i = 1
        for gene, row in df.iterrows():
            if i > num_to_keep:
                df.at[gene, 'Include in Subset'] = 'no'
            i += 1

    elif args.percent_complete:
        if args.percent_complete > 1:
            args.percent_complete = args.percent_complete / 100

        if args.out_group:
            for gene, row in df.iterrows():
                if df.at[gene, 'In-Group Completeness'] < args.percent_complete:
                    df.at[gene, 'Include in Subset'] = 'no'
        else:
            for gene, row in df.iterrows():
                if df.at[gene, 'Completeness'] < args.percent_complete:
                    df.at[gene, 'Include in Subset'] = 'no'

    df.to_csv(f'select_orthologs.tsv', sep='\t')
    return df


def update_df_taxa(df):
    """

    :return:
    """
    with open('select_taxa.tsv', 'r') as infile:
        infile.readline()
        to_drop = []
        for line in infile:
            line = line.strip()
            taxon, _, _, _, include = line.split()
            if include == 'no':
                to_drop.append(taxon)

    df = df.drop(to_drop)
    return df


if __name__ == '__main__':
    description = 'Selects orthologs to be included in super matrix construction'
    parser, optional, required = help_formatter.initialize_argparse(name='select_orthologs.py',
                                                                    desc=description,
                                                                    usage='select_orthologs.py '
                                                                          '[OPTIONS]')

    # Optional Arguments
    optional.add_argument('--out_group', type=str, metavar='out_group.txt', default=None,
                          help=textwrap.dedent("""\
                          Path to text file containing out-group taxa Unique IDs.
                          """))
    optional.add_argument('-n', '--gene_number', type=int, metavar='<N>', default=None,
                          help=textwrap.dedent("""\
                          Number of genes in subset.
                          Cannot be used with percent_complete.
                          """))
    optional.add_argument('-c', '--percent_complete', type=float, metavar='<N>', default=None,
                          help=textwrap.dedent("""\
                          Threshold for percent complete when subsetting.
                          Cannot be used with gene_number.
                          """))
    optional.add_argument('--chimeras', type=str, metavar='chimeras.tsv', default=None,
                          help=textwrap.dedent("""\
                          A TSV containing chimeras, higher and lower taxonomic designations, 
                          and the taxa comprising each chimera.
                          """))

    args = help_formatter.get_args(parser, optional, required, pre_suf=False, inp_dir=False, out_dir=False)

    config = configparser.ConfigParser()
    config.read('config.ini')
    dfo = str(Path(config['PATHS']['database_folder']).resolve())
    metadata = f'{dfo}/metadata.tsv'
    orthologs_dir = f'{dfo}/orthologs/'

    matrix = tools.completeness(args=args, input_dir=orthologs_dir, genes=True)
    if os.path.isfile('select_taxa.tsv'):
        matrix = update_df_taxa(matrix)
    taxa_count, _ = matrix.shape
    gene_comp = matrix.sum().divide(other=taxa_count)

    make_subset_tsv()

    tools.make_plot(gene_comp, f'gene_comp', y_count=taxa_count, genes=True)



#!/usr/bin/env python
import configparser
import os
import shutil
import sys
import textwrap
from pathlib import Path

from Bio import SeqIO
from Bio.Seq import Seq
from Bio.SeqRecord import SeqRecord

from phylofisher import help_formatter


def parse_ortholog_tsv():
    """
    Parses select_orthologs.tsv (output of select_orthologs.py) to determine which genes are to be included in
    the final dataset.
    :return:
    """
    with open('select_orthologs.tsv', 'r') as infile:
        header = infile.readline()
        genes_to_include = []
        for line in infile:
            line = line.strip()
            if 'In-Group Completeness' in header:
                gene, _, _, _, include = line.split('\t')
            else:
                gene, _, include = line.split('\t')
            if include == 'yes':
                genes_to_include.append(gene)

    return genes_to_include


def subset_orthologs():
    """
    Subsets orthologs from the Database into a new directory to be included in the final dataset
    :return:
    """
    # Creates output dir if it doesn't already exist
    if os.path.isdir(args.output) is False:
        os.makedirs(args.output)

    # Genes to include in the final dataset
    genes = parse_ortholog_tsv()
    # List of paths to ortholog files
    files = [os.path.join(orthologs_dir, x) for x in os.listdir(orthologs_dir) if x.endswith('.fas')]
    # Copies gene files to the output dir
    for gene in genes:
        for file in files:
            if gene == os.path.basename(file).split('.')[0]:
                src = file
                dest = f'{args.output}/{os.path.basename(file)}'
                shutil.copy(src, dest)


def parse_taxa_tsv():
    """
    Parses select_taxa.tsv (output of select_taxa.py) to determine which taxa are to be included in
    the final dataset.
    :param tsv_file:
    :return:
    """
    with open('select_taxa.tsv', 'r') as infile:
        infile.readline()
        taxa_to_include = []
        for line in infile:
            line = line.strip()
            taxon, _, _, _, include = line.split('\t')
            if include == 'yes':
                taxa_to_include.append(taxon)

    return taxa_to_include


def get_chimeras():
    """
    Parses chimeras.tsv if the chimera flag is utilized
    :return: Dict with chimera's Unique ID as the key and a list the Unique IDs of taxa comprising the chimera as the
    value
    """
    chim_dict = dict()
    chim_taxa = []
    with open(args.chimeras, 'r') as infile:
        for line in infile:
            split_line = line.strip().split('\t')
            chim_dict[split_line[0]] = split_line[3:]
            chim_taxa += split_line[3:]
    return chim_dict, chim_taxa


def subset_taxa():
    """
    Subsets taxa from the Database into a new directory to be included in the final dataset
    """
    taxa = parse_taxa_tsv()

    if args.chimeras:
        chimeras, chimeric_taxa = get_chimeras()

    files = [file for file in os.listdir(args.output)]

    for file in files:
        with open(f'{args.output}/{file}', 'r') as infile, open(f'{args.output}/tmp', 'w') as outfile:
            records = []
            if args.chimeras:
                chimera_seqs = {k: '' for k in chimeras.keys()}

            for record in SeqIO.parse(infile, 'fasta'):
                # Only includes taxa marked "yes" in select_taxa.tsv

                if record.description in taxa:
                    records.append(record)

                elif args.chimeras and record.description in chimeric_taxa:
                    for key in chimeras.keys():
                        if record.description in chimeras[key] and len(record.seq) > len(chimera_seqs[key]):
                            chimera_seqs[key] = str(record.seq)


            if args.chimeras:
                for org, seq in chimera_seqs.items():
                    if len(seq) > 0:
                        records.append(SeqRecord(Seq(seq), id=org, name='', description=''))

            SeqIO.write(records, outfile, 'fasta')

        shutil.move(f'{args.output}/tmp', f'{args.output}/{file}')


def check_if_empty():
    """
    Checks to see if there are empty files in the output directory
    :return: Boolian value (True if no empty files)
    """
    for subdir, dirs, files in os.walk(args.output):
        for file in files:
            filepath = subdir + os.sep + file

            if os.path.getsize(filepath) == 0:
                sys.exit('There are empty files in the output. Please check for errors')

    return True


if __name__ == '__main__':
    description = 'Subsets taxa and orthologs to be included in super matrix construction'
    parser, optional, required = help_formatter.initialize_argparse(name='prep_final_dataset.py',
                                                                    desc=description,
                                                                    usage='prep_final_dataset.py '
                                                                          '[OPTIONS]')

    # Optional Arguments
    optional.add_argument('--chimeras', type=str, metavar='chimeras.tsv', default=None,
                          help=textwrap.dedent("""\
                          Path to chimeras.tsv. This will collapses taxa listed in chimera.tsv into a chimera 
                          keeping the longest sequence for each gene.
                          """))

    args = help_formatter.get_args(parser, optional, required, inp_dir=False, pre_suf=False)

    config = configparser.ConfigParser()
    config.read('config.ini')
    dfo = str(Path(config['PATHS']['database_folder']).resolve())
    orthologs_dir = f'{dfo}/orthologs'

    if os.path.isdir(args.output) is False:
        os.mkdir(args.output)

    if os.path.isfile('select_orthologs.tsv'):
        subset_orthologs()
    else:
        files = [os.path.basename(file) for file in os.listdir(orthologs_dir)]
        for file in files:
            shutil.copy(f'{orthologs_dir}/{file}', f'{args.output}/{file}')

    if os.path.isfile('select_taxa.tsv'):
        subset_taxa()

    check_if_empty()



#!/usr/bin/env python
import configparser
import glob
import os
import sys
import textwrap
from collections import defaultdict
from pathlib import Path

import numpy as np
import pandas as pd
from Bio import SeqIO

from phylofisher import help_formatter, tools


def collect_names(files):
    """
    Collect all short names. Collest information about 'route'
    for newly added sequences (BBH, SBH, HMM).
    input: fasta files with genes (from fisher)
    return: set of sorted shortnames; dictionary with genes 
    as a keys and deeper level dictionaries with shortnames of organisms 
    as keys and BBH or SBH or HMM as values
    """
    names = set()
    routes = defaultdict(dict)
    for file in files:
        gene_name = file.split('/')[-1].split('.')[0]
        for record in SeqIO.parse(file, 'fasta'):
            if record.name.count('_') >= 3:
                name = record.name.split('_')[0]
                route = record.name.split('_')[2]
                routes[gene_name][name] = route
            elif '_' in record.name:
                name = record.name.split('_')[0]
            else:
                name = record.name
            names.add(name)
    return sorted(list(names)), routes


def get_gene_column(gene, names):
    """
    Collect information about number of gene variants for all organisms.
    input: fasta file for a given gene
    return:  pd.Series with information about present(>0)/absence(0) of a gene
    for all organisms. 
    example: Albugo: 3, Naegr: 0, ...
    """
    gene_name = gene.split('/')[-1].split('.')[0]
    column = pd.Series(np.zeros(len(names)), index=names, name=gene_name)
    for record in SeqIO.parse(gene, "fasta"):
        if "_" in record.name:
            org = record.name.split('_')[0]
        else:
            org = record.name
        column[org] += 1
    return column.astype(int)


def make_table(folder):
    """
    Collect information about all genes as pd.Series and organize them
    into pd.DataFrame
    input: fodler with genes in fasta format
    return: pd.DataFrame with genes; dictionary with information about
    'routes' used for sequence selection (BBH, SBH, HMM)
    """
    genes = [gene for gene in glob.glob(f'{folder}/*.fas') if os.path.isfile(gene)]
    names, routes = collect_names(genes)
    columns = []
    for gene in genes:
        columns.append(get_gene_column(gene, names))
    df = pd.DataFrame(columns)
    df = df.transpose()
    df = df.reindex(sorted(df.columns), axis=1)  # comment me
    return df, routes


def table_with_routes(df, routes):
    """Create occupancy table and add information about routes used for sequence
    selection (BBH, SBH, HMM) to input dataframe
    input: pd.DataFrame with information about all genes (present(>0)/absent(0)) 
    for all organisms, dictionary with routes for all sequences
    output: modified input dataframe with information about route
    """
    full_names = []
    high_tax_list = []
    low_tax_list = []
    for org in in_taxa_dict.keys():
        group, subtax, long_name = in_taxa_dict[org]
        high_tax_list.append(group)
        low_tax_list.append(subtax)
        full_names.append(long_name)

    df = df[df.index.isin(in_taxa_dict.keys())]
    no_seqs = set(in_taxa_dict.keys()) - set(df.index)

    for taxon in no_seqs:
        df.loc[taxon] = len(df.columns) * [0]

    df.index.name = 'Unique ID'
    df.insert(loc=0, column='Lower Taxonomy', value=low_tax_list)
    df.insert(loc=0, column='Higher Taxonomy', value=high_tax_list)
    df.insert(loc=0, column='Full Name', value=full_names)

    df = df.sort_index(axis=0)
    df.to_csv(f'{output_fold}/occupancy.tsv', sep='\t')

    # Adds routes to df
    for gene in df.columns:
        df[gene] = df[gene].apply(str)
        for org in df[gene].index:
            if org in routes[gene]:
                df.at[org, gene] = f'{df[gene][org]}_{routes[gene][org]}'

    df.to_csv(f'{output_fold}/occupancy_with_routes.tsv', sep='\t')

    return df


def check_paralogs():
    """Collect all Unique IDs for organisms with at least one paralog
    in the dataset.
    input: None
    return: set of short names of organisms with at least one paralog
    """
    paralogs = set()
    paralog_fold = os.path.dirname(args.metadata)
    for file in glob.glob(f'{paralog_fold}/paralogs/*.fas'):
        for record in SeqIO.parse(file, 'fasta'):
            paralogs.add(record.name.split('.')[0])
    return paralogs


def get_routes():
    """

    :return:
    """
    my_routes = dict()
    for org in in_taxa_dict.keys():
        sbh = {'SBH': 0, 'BBH': 0, 'HMM': 0}
        for val in res.loc[org].values[2:]:
            if '_' in val:
                count, route = val.split('_')
                sbh[route] += int(count)

        my_routes[org] = [sbh['SBH'], sbh['BBH'], sbh['HMM']]

    return my_routes


def stats_orgs(df, new_data=False):
    """
    Create tsv file with basic summary about analyzed dataset without information
    about 'routes' and paralogs.
    input: dataframe
    return: None
    """
    rows = []

    if new_data:
        df = df[df.index.isin(in_taxa_dict.keys())]
    else:
        df = df[df.index.isin(db_taxa_dict.keys())]

    df2 = df.copy()
    df2[df2 >= 1] = 1

    df = df.sum(axis=1).to_frame()

    if new_data:
        df[f"Genes out of {len(matrix.columns)}"] = df2.sum(axis=1).to_frame()
        df = df.rename(columns={0: f"Sequences Collected"})

    else:
        df = df.rename(columns={0: f"Genes out of {len(matrix.columns)}"})

    # Fill in taxonomic information
    if new_data:
        list_of_dicts = [{key: value[i] for key, value in in_taxa_dict.items()} for i in range(3)]
    else:
        list_of_dicts = [{key: value[i] for key, value in db_taxa_dict.items()} for i in range(3)]
    df['Long Name'] = df.index.map(list_of_dicts[2])
    df['Higher Taxonomy'] = df.index.map(list_of_dicts[0])
    df['Lower Taxonomy'] = df.index.map(list_of_dicts[1])

    # Rearrange Columns to Put Genes after taxa stats
    cols = df.columns.tolist()
    cols = cols[2:] + cols[:2]
    df = df[cols]

    if new_data:
        routes_dict = get_routes()
        list_of_routes_dicts = [{key: value[i] for key, value in routes_dict.items()} for i in range(3)]
        df["#SBH"] = df.index.map(list_of_routes_dicts[0])
        df["#BBH"] = df.index.map(list_of_routes_dicts[1])
        df["#HMM"] = df.index.map(list_of_routes_dicts[2])
        out_filename = 'new_taxa_stats.tsv'
    else:
        out_filename = 'db_taxa_stats.tsv'

    # Fill in columns for including in SGT construction. By default all are yes
    has_paralogs = check_paralogs()
    if new_data:
        sgt_dict = {org: 'yes' for org in in_taxa_dict.keys()}
    else:
        sgt_dict = {org: 'yes' for org in db_taxa_dict.keys()}
    df['SGT'] = df.index.map(sgt_dict)

    # Fill in column for paralogs. If no paralogs entry is 'none'.
    # If there are paralogs entry is 'yes'. If there are paralogs, but --ortholog_only is given entry is 'no'.
    if new_data:
        pass
    else:
        paralogs_dict = {org: ('yes' if org in has_paralogs and not args.orthologs_only
                               else 'no' if org in has_paralogs and args.orthologs_only else 'none')
                         for org in db_taxa_dict}
        df['Paralogs'] = df.index.map(paralogs_dict)

    df = df.rename_axis('Unique ID')
    df.to_csv(f'{output_fold}/{out_filename}', sep='\t')


def stats_gene(df):
    """Write basic summary for all genes to a file
    input: dataframe
    return: None
    """
    # res.write(f'Gene,Total,total[%],SGT\n')
    taxa_count = len(df)
    df[df != 0] = 1
    df = df.sum().to_frame()
    df = df.rename(columns={0: 'Number of Taxa'})
    df[f'Percent of Total Taxa (out of {taxa_count})'] = round((df['Number of Taxa'] / taxa_count) * 100, 2)
    df = df.rename_axis('Gene Name')
    df = df.sort_values(by=['Number of Taxa'], ascending=False)
    df['SGT'] = ['yes'] * len(df)
    df.to_csv(f'{output_fold}/gene_stats.tsv', sep='\t')


if __name__ == '__main__':
    description = 'Produce preliminary statistics about newly input data.'
    parser, optional, required = help_formatter.initialize_argparse(name='informant.py',
                                                                    desc=description,
                                                                    usage='informant.py -i input_folder [OPTIONS]')

    # Optional Arguments
    optional.add_argument('--orthologs_only', action='store_true',
                          help=textwrap.dedent("""\
                          Paralogs will NOT be included from any taxa in the starting 
                          database in downstream single gene tree construction.
                          """))

    in_help = 'Path to fisher.py output directory'
    args = help_formatter.get_args(parser, optional, required, pre_suf=False, in_help=in_help, out_dir=False)

    config = configparser.ConfigParser()
    config.read('config.ini')
    dfo = str(Path(config['PATHS']['database_folder']).resolve())
    args.input_metadata = os.path.abspath(config['PATHS']['input_file'])
    args.metadata = os.path.join(dfo, 'metadata.tsv')

    output_fold = os.path.basename(os.path.normpath(args.input)) + '/informant_stats'
    if os.path.isdir(output_fold):
        sys.exit(f'Error: {output_fold} folder already exists.')
    else:
        os.mkdir(output_fold)

    db_taxa_dict = tools.parse_metadata(args.metadata)
    in_taxa_dict = tools.parse_metadata(args.input_metadata, input_meta=True)
    all_taxa_dict = {**db_taxa_dict, **in_taxa_dict}

    matrix, routes = make_table(args.input)

    res = table_with_routes(matrix, routes)

    stats_orgs(matrix)
    stats_orgs(matrix, new_data=True)
    stats_gene(matrix)



#!/usr/bin/env python
import configparser
import csv
import os
import shutil
import sys
import textwrap
from glob import glob
from pathlib import Path

from Bio import SeqIO

from phylofisher import help_formatter, tools


def parse_metadata():
    meta = os.path.join(dfo, 'metadata.tsv')
    with open(meta, 'r') as f:
        reader = csv.reader(f, delimiter='\t')
        lines = list(reader)
        return lines


def parse_input():
    file_contents = set()
    with open(args.input, 'r') as infile:
        for line in infile:
            line = line.strip()
            file_contents.add(line)

    return list(file_contents)


def check_metadata():
    for item in parse_input():
        in_meta = False
        for line in parse_metadata():
            if item in line:
                in_meta = True

        if not in_meta:
            sys.exit(f'{item} is not in the database. Please check your input file.')



def fasta_cleaner(file, org_set):
    records = list(SeqIO.parse(file, 'fasta'))
    with open(file, 'w') as res:
        for record in records:
            if record.name.split('.')[0] not in org_set:
                res.write(f'>{record.name}\n{record.seq}\n')


def delete_homologs(org_set):
    for folder in ['orthologs', 'paralogs']:
        files = glob(os.path.join(dfo, folder) + '/*.fas')
        for file in files:
            fasta_cleaner(file, org_set)


def delete_proteomes(org_set):
    for org in org_set:
        os.remove(os.path.join(dfo, 'proteomes', f'{org}.faa.tar.gz'))


def purge():
    to_remove = parse_input()
    meta = os.path.join(dfo, 'metadata.tsv')

    lines = parse_metadata()
    orgs_to_del = set()
    
    with open(meta, 'w') as out_file:
        res = csv.writer(out_file, delimiter='\t')
        for line in lines:
            if line[2] in to_remove:
                orgs_to_del.add(line[0])
            elif line[3] in to_remove:
                orgs_to_del.add(line[0])
            elif line[0] in to_remove:
                orgs_to_del.add(line[0])
            else:
                res.writerow(line)
                
    delete_homologs(orgs_to_del)
    delete_proteomes(orgs_to_del)


# TODO input as a file
if __name__ == '__main__':
    description = 'Deletes taxa and/or taxonomic groups from the database'
    parser, optional, required = help_formatter.initialize_argparse(name='purge.py',
                                                                    desc=description,
                                                                    usage='purge.py [OPTIONS] -i to_purge.txt -d path/to/database')

    # Optional Arguments
    required.add_argument('-i', '--input', type=str, metavar='to_purge.txt',
                          help=textwrap.dedent("""\
                          Path to text file containing Unique IDs and Taxonomic designations of organisms for deletion.
                           """))
    required.add_argument('-d', '--database', metavar='<input_dir>', type=str,
                          help=textwrap.dedent("""\
                          Path to database to purge.
                          """))

    in_help = 'Path to database directory'
    args = help_formatter.get_args(parser, optional, required, pre_suf=False, inp_dir=False, out_dir=False)

    dfo = os.path.abspath(args.database)

    check_metadata()
    tools.backup(dfo)
    purge()



#!/usr/bin/env python
import os
import shutil
import tarfile
import textwrap
from datetime import datetime

from phylofisher import help_formatter


def list_backups():
    """
    List available backups in the database/backups dir
    :return:
    """
    print('Available Backups to Restore From:\n')
    print(f'Backup Num:\tDate:\t\tTime:')
    for i, date_time in enumerate(backups):
        date_time_str = datetime.strftime(date_time, "%d-%b-%Y_%H-%M-%S")
        date = date_time_str.split('_')[0]
        time = date_time_str.split('_')[1]
        print(f'{i + 1}\t\t{date.replace("-"," ")}\t{time.replace("-",":")}')


def restore():
    """
    Restores database from user specified backup
    :return:
    """
    backup_dir = f'{args.database}/backups/{datetime.strftime(backups[args.restore - 1], "%d-%b-%Y_%H-%M-%S")}'

    if os.path.isdir(f'{args.database}/orthologs'):
        shutil.rmtree(f'{args.database}/orthologs')
    shutil.copytree(f'{backup_dir}/orthologs', f'{args.database}/orthologs')

    if os.path.isdir(f'{args.database}/paralogs'):
        shutil.rmtree(f'{args.database}/paralogs')
    shutil.copytree(f'{backup_dir}/paralogs', f'{args.database}/paralogs')

    if os.path.isfile(f'{args.database}/metadata.tsv'):
        os.remove(f'{args.database}/metadata.tsv')
    shutil.copy(f'{backup_dir}/metadata.tsv', f'{args.database}/metadata.tsv')

    if os.path.isfile(f'{args.database}/tree_colors.tsv'):
        os.remove(f'{args.database}/tree_colors.tsv')
    shutil.copy(f'{backup_dir}/tree_colors.tsv', f'{args.database}/tree_colors.tsv')

    if os.path.isdir(f'{args.database}/proteomes'):
        shutil.rmtree(f'{args.database}/proteomes')
    shutil.copytree(f'{backup_dir}/proteomes', f'{args.database}/proteomes')


if __name__ == '__main__':
    desc = 'Utility to restore the PhyloFisher Database from backups'
    parser, optional, required = help_formatter.initialize_argparse(name='backup_restoration.py',
                                                                    desc=desc,
                                                                    usage="backup_restoration.py -d <db_dir>")

    required.add_argument('-d', '--database', type=str, metavar='<db_dir>', required=True,
                          help=textwrap.dedent("""\
                          Path to database directory.
                          """))

    optional.add_argument('--list_backups', action='store_true',
                          help=textwrap.dedent("""\
                          List available backups to restore from.
                          """))
    optional.add_argument('--restore', type=int, metavar='N',
                          help=textwrap.dedent("""\
                          Backup number to restore from. Backup number is obtained through the --list_backups option.
                          """))

    args = help_formatter.get_args(parser, optional, required, out_dir=False, pre_suf=False, inp_dir=False)

    backup_dir = f'{os.path.abspath(args.database)}/backups'
    backups = os.listdir(backup_dir)
    backups = [datetime.strptime(date, '%d-%b-%Y_%H-%M-%S') for date in backups]
    backups.sort()

    if args.list_backups:
        list_backups()

    if args.restore:
        restore()



#!/usr/bin/env python
import glob
import math
import os
import random
import shutil
import tarfile
import textwrap

import numpy
import pandas as pd

from phylofisher import help_formatter


def make_tmp_dir(base):
    """
    Makes tmpo directory. If it exists the old tmp directory is first removed
    """
    tmp_dir = f'{base}/tmp'
    try:
        os.mkdir(tmp_dir)
    except FileExistsError:
        shutil.rmtree(tmp_dir)
        os.mkdir(tmp_dir)


def get_num_of_replicates(x):
    """
    Uses 0.95=1-(1-x/100)^n to determine the number of replicates where x is the percentage of genes samples and
    n is the number of replicates
    """
    bottom_eq = 1 - (x / 100)
    confidence = 1 - args.confidence_interval
    n = numpy.log(confidence) / numpy.log(bottom_eq)
    n = math.ceil(n)
    return n


def subsample(percent, genes, output):
    """
    Creates gene subsets and super matirices of the subsampled gene sets.
    """
    replicate_num = get_num_of_replicates(percent)
    # Creates gene subsample sets with the number of replicates calculated in get_num_of_replicates
    for i in range(replicate_num):
        size = len(genes) * percent * 0.01
        sample = random.sample(genes, int(size))
        make_tmp_dir(output)
        # copy gene files into a tmp directory
        for file in sample:
            shutil.copy(file, f'{output}/tmp')
        # Runs forge to create super matrix of subsampled gene sets
        os.system(f'matrix_constructor.py '
                  f'-i {output}/tmp '
                  f'-o {output}/replicate_{i + 1} '
                  f'-if {args.in_format.lower()} '
                  f'-of {args.out_format.lower()} '
                  f'-c')
        shutil.rmtree(f'{output}/tmp')


def make_csv():
    perc_dict = dict()
    for x in sampling_percents:
        perc_dict[str(x)] = dict()
    for root, dirs, files in os.walk(args.output):
        for file in files:
            if file == 'indices.tsv':
                path = os.path.join(root, file)
                _, perc, rep, _ = path.split('/')
                perc, _ = perc.split('_')
                _, rep = rep.split('_')

                with open(path, 'r') as infile:
                    perc_dict[str(perc)][rep] = []
                    infile.readline()
                    for line in infile:
                        line = line.strip()
                        gene, _, _ = line.split('\t')
                        perc_dict[str(perc)][rep].append(gene)

    for percent in perc_dict.keys():
        data = perc_dict[percent]
        data = {int(k): v for k, v in data.items()}
        df = pd.DataFrame(data).transpose()
        df = df.sort_index().transpose()
        df = df.add_prefix('Replicate ')

        df.to_csv(f'{args.output}/{percent}_Percent.tsv', index=False, sep='\t')


def clean_up():
    """
    Copies matricies in to root output directory and tar's "percent" directories
    """
    for root, dirs, files in os.walk(args.output):
        for file in files:
            if file.endswith('.fas') or file.endswith('.phy'):
                path = os.path.join(root, file)
                ext = path.split('.')[-1]
                _, perc, rep, _ = path.split('/')
                perc, _ = perc.split('_')
                _, rep = rep.split('_')

                new_name = f'{args.output}/{perc}rep{rep}.{ext}'
                shutil.copy(path, new_name)

    os.chdir(args.output)
    for root, dirs, files in os.walk('.'):
        for dir in dirs:
            if dir.endswith('Percent'):
                with tarfile.open(f'{dir}.tar.gz', mode='x:gz') as my_tar:
                    my_tar.add(dir)
                shutil.rmtree(dir)


if __name__ == '__main__':
    description = 'Constructs super-matrices from randomly sampled genes.'
    parser, optional, required = help_formatter.initialize_argparse(name='random_resampler.py',
                                                                    desc=description,
                                                                    usage='random_resampler.py '
                                                                          '[OPTIONS] -i /path/to/input/')

    # Optional Arguments
    optional.add_argument('-if', '--in_format', metavar='<format>', type=str, default='fasta',
                          help=textwrap.dedent("""\
                              Format of the input single gene alignments.
                              Options: fasta, phylip (names truncated at 10 characters), 
                              phylip-relaxed (names are not truncated), or nexus.
                              Default: fasta
                              """))
    optional.add_argument('-of', '--out_format', metavar='<format>', type=str, default='fasta',
                          help=textwrap.dedent("""\
                              Desired format of the output steps.
                              Options: fasta, nexus, phylip (names truncated at 10 characters), 
                              or phylip-relaxed (names are not truncated)
                              Default: fasta
                              """))
    optional.add_argument('-ci', '--confidence_interval', metavar='0.N', type=float, default=0.95,
                          help=textwrap.dedent("""\
                              Confidence interval to use to calculate the number of replicates required.
                              Default: 0.95%%
                              """))
    optional.add_argument('-ps', '--percent_sampling', metavar='N', type=int, default=20,
                          help=textwrap.dedent("""\
                              Percent sampling step size.
                              Default: 20%% (e.g. 5%%, 10%%, 20%%, 25%%)
                              The default 20%% sampling results in a sampling series of 20%%, 40%%, 60%%, and 80%%.
                              """))

    in_help = 'Path to input directory containing gene files in FASTA format.'
    args = help_formatter.get_args(parser, optional, required, in_help=in_help)

    args.output = f'{args.output}_{args.percent_sampling}ps_{args.confidence_interval}ci'

    os.mkdir(args.output)
    # creates a list of percents to sample gene set at
    sampling_percents = list(range(0, 100, args.percent_sampling))[1:]
    # gets list of all gene files in the input directory
    files = [file for file in glob.glob(f'{args.input}/{args.prefix}*{args.suffix}')]

    # iterated through list of sampling percentages and creates a directory to store the outputs
    for x in sampling_percents:
        new_dir = f'{args.output}/{str(x)}_Percent'
        os.mkdir(new_dir)
        subsample(x, files, new_dir)

    make_csv()
    clean_up()



#!/usr/bin/env python
import os
import sys
import textwrap

import matplotlib.pyplot as plt
import pandas as pd
import scipy.cluster.hierarchy as shc
from Bio import SeqIO, Phylo
from Bio.SeqUtils.ProtParam import ProteinAnalysis
from ete3 import Tree, NodeStyle, TreeStyle, TextFace, CircleFace, faces, AttrFace

from phylofisher import help_formatter


def distance_matrix2tree(Z, names):
    """Return tree representation for distance matrix"""
    n = Z.shape[0] + 1
    i2n = [0] * (2 * n - 1)
    t = Tree()
    for i, (idx1, idx2, dist, sample_count) in enumerate(Z):
        idx1, idx2 = int(idx1), int(idx2)
        # create Tree object for tips / leaves
        if idx1 < n:
            i2n[idx1] = Tree(name=names[idx1])
        if idx2 < n:
            i2n[idx2] = Tree(name=names[idx2])
        # create new node
        t = Tree()
        # normalise distance
        dist1 = dist - i2n[idx1].get_farthest_leaf()[1]
        dist2 = dist - i2n[idx2].get_farthest_leaf()[1]
        # add children
        t.add_child(i2n[idx1], dist=dist1)
        t.add_child(i2n[idx2], dist=dist2)
        # store
        i2n[n + i] = t
    return t


def aa_comp_calc():
    peptides = ['A', 'G', 'P', 'S', 'T', 'C', 'F', 'W', 'Y', 'H', 'R', 'K', 'M', 'I', 'L', 'V', 'N', 'D', 'E', 'Q']
    if not os.path.isdir(args.output):
        os.mkdir(args.output)
    with open(args.input, 'r') as infile, open(f'{args.output}/aa_comp.tsv', 'w') as outfile:
        outfile.write('Taxon\t' + '\t'.join(peptides) + '\n')

        # Reads in input file
        for record in SeqIO.parse(infile, format=args.in_format):
            outfile.write(f'{record.id}\t')
            analysed_seq = ProteinAnalysis(str(record.seq))
            count_dict = analysed_seq.count_amino_acids()
            length = len(str(record.seq).replace("-", "").replace("X", "").replace("*", ""))
            out_str = ''

            # Loops through peptides and checks to see if it is in count_dict
            for pep in peptides:
                if pep in count_dict.keys():
                    out_str += f'{float(count_dict[pep]) / length}\t'
                else:
                    out_str += '0\t'

            outfile.write(out_str.strip() + '\n')


def parse_tree_colors():
    color_dict = {}
    tree_color_dict = {}
    meta_dict = {}

    with open(tree_colors, 'r') as infile:
        infile.readline()
        for line in infile:
            s_line = line.strip().split('\t')
            tree_color_dict[s_line[0]] = s_line[1]

    with open(metadata, 'r') as infile:
        infile.readline()
        for line in infile:
            s_line = line.strip().split('\t')
            meta_dict[s_line[0]] = s_line[2:4]

    for k, v in meta_dict.items():
        if k in tree_color_dict.keys():
            color_dict[k] = tree_color_dict[k]
        elif v[0] in tree_color_dict.keys():
            color_dict[k] = tree_color_dict[v[0]]
        elif v[1] in tree_color_dict.keys():
            color_dict[k] = tree_color_dict[v[1]]
        else:
            sys.exit(f'{k} is not in the tree_color.tsv')

    return color_dict, tree_color_dict, meta_dict


def mylayout(node):
    # Internal Node formatting
    nstyle = NodeStyle()
    nstyle["size"] = 0
    nstyle["vt_line_width"] = 2
    nstyle["hz_line_width"] = 2

    if node.is_leaf():
        circle_face = CircleFace(10, color_dict[node.name], style='circle')
        faces.add_face_to_node(circle_face, node, position="branch-right", column=0)

        text_face = TextFace(meta_dict[node.name][0])
        text_face.fgcolor = color_dict[node.name]
        faces.add_face_to_node(text_face, node, position="branch-right", column=1)

        node.set_style(nstyle)

    elif node.is_root():
        node.dist = 0
        node.set_style(nstyle)

    else:
        node.set_style(nstyle)


def make_plot():
    df = pd.read_csv(f'{args.output}/aa_comp.tsv', sep="\t")
    df = df.set_index('Taxon')

    z = shc.linkage(df, method='ward')
    t = distance_matrix2tree(z, df.index.values)
    t.write(outfile=f'{args.output}/distance_matrix.tre')

    # Basic tree style
    ts = TreeStyle()
    ts.show_leaf_name = False
    ts.mode = "c"
    ts.show_scale = False
    ts.layout_fn = mylayout
    # PDF rendering
    t.render(f'{args.output}/aa_comp_calculator.pdf', w=183, units="mm", tree_style=ts)


if __name__ == '__main__':
    description = 'Calculates amino acid composition of the supplied super matrix'
    parser, optional, required = help_formatter.initialize_argparse(name='aa_comp_calculator.py',
                                                                    desc=description,
                                                                    usage='aa_comp_calculator.py '
                                                                          '[OPTIONS] -i <matrix>')

    required.add_argument('-i', '--input', required=True, type=str, metavar='matrix',
                          help=textwrap.dedent("""\
                          Path to input matrix for analysis.
                          """))
    required.add_argument('-d', '--database', type=str, metavar='<db_dir>', required=True,
                          help=textwrap.dedent("""\
                          Path to database directory.
                          """))
    optional.add_argument('-if', '--in_format', metavar='<format>', type=str, default='fasta',
                          help=textwrap.dedent("""\
                          Format of input matrix.
                          Options: fasta, nexus, phylip (names truncated at 10 characters), 
                          Default: fasta
                          """))

    args = help_formatter.get_args(parser, optional, required, pre_suf=False, inp_dir=False)

    dfo = os.path.abspath(args.database)

    tree_colors = f'{dfo}/tree_colors.tsv'
    metadata = f'{dfo}/metadata.tsv'

    color_dict, tree_color_dict, meta_dict = parse_tree_colors()

    aa_comp_calc()
    make_plot()



#!/usr/bin/env python
import sys
import csv
import configparser
import textwrap
from pathlib import Path
from phylofisher import help_formatter


def check_input_meta(dataset_fold, input_meta):
    """
    Checks that all ids are unique in input_metadata and also between
    input_metadata and metadata. 
    Checks for illegal characters in input ids.
    """

    database_metadata_ids = set()
    with open(str(Path(dataset_fold, 'metadata.tsv')), 'r') as csvfile:
        reader = csv.reader(csvfile, delimiter='\t')
        for row in reader:
            id_ = row[0]
            if id_ != 'Unique ID':
                database_metadata_ids.add(id_)

    input_ids = set()
    with open(input_meta, 'r') as csvfile:
        reader = csv.reader(csvfile, delimiter='\t')
        for row in reader:
            id_ = row[2]
            if id_ == 'Unique ID':
                continue
            
            #check that ID is unique
            if id_ in input_ids:
                print("ERROR: ", id_, "is not a unique ID")
                sys.exit()

            if id_ in database_metadata_ids:
                print("ERROR: ", id_, "already in the database")
                sys.exit()

            #check illegal characters
            for ch in ["_", '@', '..', '*', ' ']:
                if ch in id_:
                    print(f"ERROR: illegal character {ch} in {id_}")
                    sys.exit()

            input_ids.add(id_)


if __name__ == '__main__':
    description = 'Script for the analysis folder configuration.'
    usage = 'config.py -d <database_folder> -i <input_file.tsv> [OPTIONS]'
    parser, optional, required = help_formatter.initialize_argparse(name='config.py',
                                                                    desc=description,
                                                                    usage=usage)

    # Required Arguments
    required.add_argument('-d', '--database_folder', required=True, metavar='<database_dir>',
                          help=textwrap.dedent("""\
                          Path to directory containing the PhyloFisher database.
                          """))
    required.add_argument('-i', '--input_file', required=True, metavar='<input.tsv>',
                          help=textwrap.dedent("""\
                          Path to input metadata in tsv format.
                          """))
    # Optional Arguments
    optional.add_argument('--orthomcl', metavar='<omcl_data>',
                          help=textwrap.dedent("""\
                          Path to orthomcl if NOT in PhyloFisherDatabase_v1.0/database/orthomcl
                          """))
    optional.add_argument('--tree_colors', metavar='<omcl_data>',
                          help=textwrap.dedent("""\
                              Path to alternative single gene tree color configuration file.
                              """))

    config = configparser.ConfigParser()
    args = help_formatter.get_args(parser, optional, required, pre_suf=False, inp_dir=False, out_dir=False)


    check_input_meta(args.database_folder, args.input_file)

    if not args.orthomcl:
        args.orthomcl = str(Path(args.database_folder, 'orthomcl'))

    if not args.tree_colors:
        args.tree_colors = str(Path(args.database_folder, 'tree_colors.tsv'))

    with open('config.ini', 'w') as configfile:
        config['PATHS'] = {'database_folder': args.database_folder,
                           'input_file':     args.input_file,
                           'orthomcl':       args.orthomcl,
                           'color_conf':     args.tree_colors}
        config.write(configfile)



#!/usr/bin/env python
import configparser
import os
import subprocess
import sys
import textwrap
import warnings
from collections import defaultdict
from functools import partial
from multiprocessing import Pool
from pathlib import Path
from shutil import copyfile, rmtree

from Bio import BiopythonExperimentalWarning
from Bio import SeqIO
from ete3 import Tree

from phylofisher import help_formatter

with warnings.catch_warnings():
    warnings.simplefilter('ignore', BiopythonExperimentalWarning)
    from Bio import SearchIO


class Hit:
    # TODO: it would be nice to refractor me
    """ Represents candidate hit.

    attributes:
    - query: refers to a gene
    - path_: refers to 'SBH' - specific query best hit
                    'BBH' - best blast hit (doesnt cluster with
                            expected taxonomical group)
                    'HMM' - selected just according to Hmmsearch
    - quality: q1,q2...qn refers to position in blast/hmmsearch
                quality also contains 'r' or 'n' after number
                which mean corresponding and non_corresponding best blast hit
                example: q1r means best blast/hmmer hit (depends on path_)
                        which is also reciprocal best blast hit to query
                        gene in our database"""

    def __init__(self, name, seq, query, hmm_hits):
        self.name = name
        self.seq = seq
        self.query = query
        self.hmm_hits = hmm_hits
        self.path_ = ""
        self.quality = ""

    def in_hmm_hits(self):
        """Controls if hit in also in hmmer hits"""
        if self.name in self.hmm_hits:
            return True
        return False


class Query:
    """Represents phylogeneticaly not specific query (only hmmsearch)

    attributes:
    - query: refers to a gene
    - squery: always False, not a sprcific query (fix it)"""

    def __init__(self, query, hmm_hits, infile_proteins):
        self.query = query
        self.hmm_hits = hmm_hits
        self.infile_proteins = infile_proteins
        self.squery = False

    def hits(self):
        """parse all hmmerhits as Hit objets
        result is a generator producing one hit at a time"""
        for hit in self.hmm_hits:
            hit = Hit(hit, self.infile_proteins[hit], self.query, self.hmm_hits)
            hit.path_ = "HMM"
            yield hit


class SpecQuery:
    """Represents phylogeneticaly specific query.

    attributes:
    - query: refers to a gene
    - organisms: organisms from which blast queries should be selected.
    """

    def __init__(self, query, spec_queries, hmm_hits, infile_proteins):
        self.query = query
        self.hmm_hits = hmm_hits
        self.seqs = []
        self.infile_proteins = infile_proteins
        self.organisms = [org.strip(' "\'') for org in spec_queries.split(',')]

    def get_specific_query(self):
        """This function returns seed blast sequence from selected specific queries(organisms).
        When target gene is not present in the first organism (first in the list) it
        will check next one.
        Returns: - True if at_least_one organism has a specific query gene
         and set seed sequences for blast as self.seqs
        if gene is present in self.organisms or None"""
        at_least_one = False
        gene_dict = {}
        for record in SeqIO.parse(str(Path(dfo, f'orthologs/{self.query}.fas')), 'fasta'):
            gene_dict[record.name] = str(record.seq)
        for org in self.organisms:
            if org in gene_dict:
                self.seqs.append(gene_dict[org])
                at_least_one = True
        return at_least_one

    def spec_query_blast(self):
        """Blast against sample with sequence selected according to spec queries.
        Returns: list with ordered best blast hits."""
        db = f'{args.output}/tmp/{sample_name}/{os.path.basename(infile)}.blastdb'
        qfile = f'{args.output}/tmp/{sample_name}/{self.query}_collected.fas'
        bout = f'{args.output}/tmp/{sample_name}/{self.query}.blastout'
        with open(qfile, 'w') as f:
            n = 0  # just for uniq names of spec query sequences
            for query_sequence in self.seqs:
                f.write(f'>{n}\n{query_sequence.replace("-", "")}\n')

        cmd = f'blastp -evalue 1e-10 -query {qfile} -db {db} -out {bout} -outfmt "6 qseqid sseqid evalue"'
        subprocess.run(cmd, shell=True, stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL)
        blast_hits = []
        for line_ in open(bout):
            prot_name = line_.split('\t')[1]
            if prot_name not in blast_hits:
                blast_hits.append(prot_name)
        return blast_hits

    def hits(self):
        """Generator function which return Hit objects prioritized by blast search
        when some specific query is present or prioritized by hmmsearch"""
        spec_query = self.get_specific_query()
        if spec_query:
            blast_hits = self.spec_query_blast()
            if not blast_hits:
                return None
            for blast_hit in blast_hits:
                hit = Hit(blast_hit, self.infile_proteins[blast_hit], self.query, self.hmm_hits)
                hit.path_ += "SBH"
                yield hit
        else:
            for hmm_hit in self.hmm_hits:
                hit = Hit(hmm_hit, self.infile_proteins[hmm_hit], self.query, self.hmm_hits)
                hit.path_ += "HMM"
                yield hit


def length_check(trimmed_aln):
    """Returns set of names of sequenes which have meaningful part
    (without X or -) longer than 30% of trimmed alignment."""
    correct_length = {}
    for record in SeqIO.parse(trimmed_aln, 'fasta'):
        if '@' in record.name:
            true_length = len(str(record.seq).replace('-', '').replace('X', ''))
            if true_length / len(record.seq) > 0.3:
                correct_length[record.name] = round((true_length / len(record.seq)), 2)
    return correct_length


def makeblastdb():
    """Prepares blast database from sample input file"""
    cmd = f"makeblastdb -in {infile} -out {args.output}/tmp/{sample_name}/{os.path.basename(infile)}.blastdb -dbtype prot"
    subprocess.run(cmd, shell=True, stdout=subprocess.DEVNULL)


def hmmer(query):
    """Performs hmmsearch with a given gene.

    returns: tuple with (gene_name, list of hmm hits [hmm1, hmm2 ... hmmn]"""
    hmm_prof = str(Path(dfo, f'profiles/{query}.hmm'))
    cmd = f'hmmsearch -E 1e-10 {hmm_prof} {infile} > {args.output}/tmp/{sample_name}/{query}.hmmout'
    subprocess.run(cmd, shell=True, stdout=subprocess.DEVNULL)
    hits = []
    for hit in SearchIO.read(f'{args.output}/tmp/{sample_name}/{query}.hmmout', 'hmmer3-text'):
        hits.append(hit.id)
    return query, hits


def get_gene_dict(threads, infile_proteins, spec_queries=None):
    """This function prepare SpeqQeury or Query for all genes.
    Query type depends if spec_queries are specified in input metadata
    for a given organism."""
    gene_dict = {}
    with Pool(processes=threads) as pool:
        hmm_pool = list(pool.map(hmmer, profiles))

        for query, hmm_hits in hmm_pool:
            if hmm_hits:
                if spec_queries:
                    gene_dict[query] = SpecQuery(
                        query, spec_queries, hmm_hits, infile_proteins)
                else:
                    gene_dict[query] = Query(query, hmm_hits, infile_proteins)
            else:
                print(f"{query} exluded. No hmm hits.")
    return gene_dict


def best_hits(max_hits, gene):
    """Selects candidates(hits) up to max_hits.
    Candidates have to be in hmm_hits

    returns: tuple with gene name and list with candidates names"""
    candidates = []
    hits = gene.hits()
    counter = 0
    if hits:
        for hit in hits:
            if counter == max_hits:
                break
            counter += 1
            if hit.in_hmm_hits() is True:
                candidates.append(hit)
    return gene.query, candidates


def bac_gog_db():
    """Parses bacterial names and gene: orthogroup/df from orthomcl.

    returns tuple with ls of [bacterial names] and {gene: orthogtoup/df} dir"""
    bac = set()
    for line_ in open(str(Path(dfo, 'orthomcl/bacterial'))):
        bac.add(line_[:-1])
    g_og = {}
    for line_ in open(str(Path(dfo, 'orthomcl/gene_og'))):
        sline = line_.split('\t')
        g_og[sline[0]] = sline[1]
    return bac, g_og


def get_infile_proteins():
    """Parses infile proteins.

    returns: dict {prot_name:seq}"""
    infile_proteins = {}
    for record in SeqIO.parse(infile, 'fasta'):
        infile_proteins[record.id] = str(record.seq)
    return infile_proteins


def get_candidates(threads, max_hits, queries):
    """pararellized best_hits function
    return list of tuples (gene:[candidate names],..)"""
    with Pool(processes=threads) as pool:
        func = partial(best_hits, max_hits)
        candidates = list(pool.map(func, queries))
        return candidates


def get_hmm_profiles():
    """Parses gene names.

    Returns: list with gene names"""
    hmm_profiles = []
    for file in os.listdir(str(Path(dfo, 'profiles/'))):
        hmm_profiles.append(file.split('.')[0])
    return hmm_profiles


def makedirs():
    """Creates output dictionaries tmp and fasta"""
    directories = [args.output, f'{args.output}/tmp']
    for directory in directories:
        if not os.path.isdir(directory):
            os.mkdir(directory)


def phylofisher(threads, max_hits, spec_queries=None):
    infile_proteins = get_infile_proteins()
    if spec_queries:
        # makes blastdb from input proteins
        makeblastdb()
        gene_dict = get_gene_dict(threads, infile_proteins, spec_queries)
    else:
        gene_dict = get_gene_dict(threads, infile_proteins)

    # return list of tuples (gene:[candidate names],..)
    candidates = get_candidates(threads, max_hits, gene_dict.values())
    # writes all candidates to for_diamond.fasta"
    with open(f'{args.output}/tmp/for_diamond.fasta', 'a') as f:
        for gene, candi_list in candidates:
            if candi_list:
                for hit in candi_list:
                    f.write(f'>{hit.name}_{hit.path_}@{gene}\n{hit.seq}\n')


def taxonomy_dict():
    """Parses taxonomical group for all organisms from metadata."""
    tax_g = {}
    for line_ in open(str(Path(dfo, 'metadata.tsv'))):
        if 'Full Name' not in line_:
            sline = line_.split('\t')
            tax = sline[0].strip()
            group = sline[2].strip()
            tax_g[tax] = group
    return tax_g


def cluster_rename_sequences():
    """Clusters sequences in input fasta file for given organism
    and rename every sequence with shortname_number

    returns: abs path to the file with clustered and renamed sequences"""
    clustered = f'{args.output}/tmp/{sample_name}/clustered.fasta'
    cmd = f'cd-hit -i {fasta_file} -o {clustered} -c 0.98'
    # performs cd-hit
    subprocess.run(cmd, shell=True, stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL)
    original_names = {}
    n = 1
    # prepares files with clustered and renamed sequences
    with open(f'{args.output}/tmp/{sample_name}/clustered_renamed.fasta', 'w') as res:
        for record in SeqIO.parse(clustered, 'fasta'):
            new_name = f"{sample_name}_{n}"
            original_names[new_name] = record.name
            res.write(f'>{new_name}\n{record.seq}\n')
            n += 1
    abs_path = os.path.abspath(f'{args.output}/tmp/{sample_name}/clustered_renamed.fasta')

    # prepares file tsv files which keep information about old name: new name
    with open(f'{args.output}/original_names.tsv', 'a') as f:
        for new_name, or_name in original_names.items():
            f.write(f'{or_name}\t{new_name}\n')
    return abs_path


def is_aa_seq(file):
    """check input gene files to make sure they are amino acid sequences and not nucletotide sequences"""
    with open(file, 'r') as infile:
        all_count, atcg_count = 0, 0
        for line in infile:
            if line.startswith(">"):
                pass
            else:
                all_count += len(line)
                atcg_count += (line.upper().count('A') + line.upper().count('T')
                               + line.upper().count('C') + line.upper().count('G'))
        atcg_per = atcg_count / all_count
        if atcg_per > 0.7:
            return False
        else:
            return True


def check_input():
    """Checks input metadata file for possible mistakes."""
    errors = ''
    taxonomic_groups = set(tax_group.values())
    n = 1
    for line in open(input_metadata):
        line = line.strip()
        if "File Name" not in line:
            n += 1
            metadata_input = line.split('\t')
            f_file = str(Path(metadata_input[0], metadata_input[1]))
            # checks if file for given organism exists
            if os.path.isfile(f_file.strip()) is False:
                errors += f"line {n}: file {f_file} doesn't exist\n"
            # Checks if file is amino acid seq or nucleotide seq
            elif is_aa_seq(f_file) is False:
                errors += f"line {n}: file {f_file} contains nucleotide sequences\n"
            s_name = metadata_input[2].strip()
            if s_name in tax_group:
                # checks if short name is already included in metadata
                # from the database
                errors += f'line {n}: {s_name} already in metadata\n'
            s_queries = metadata_input[5]
            if s_queries.lower().strip() != 'none':
                for q in s_queries.split(','):
                    if q.strip() not in tax_group:
                        # checks if specific query exists in metadata
                        errors += f'line {n}: {q} not in metadata\n'
                tax = metadata_input[3].strip()
                if '*' not in tax:
                    # checks if taxonomic group is already in metadata.
                    # If you want to use new tax group: use * in its name
                    if tax not in taxonomic_groups:
                        errors += f'line {n}: {tax} not in metadata\n'
    if errors:
        sys.exit(f'Please check your input file:\n{errors[:-1]}')


def diamond():
    """executes diamond against orthomcl and datasetdb database"""
    for_diamond = f'{args.output}/tmp/for_diamond.fasta'

    db = str(Path(dfo, 'orthomcl/orthomcl.diamonddb'))
    out = f'{args.output}/tmp/orthomcl_diamond.res'
    cmd = (
        f'diamond blastp -e 1e-10 -q {for_diamond} --more-sensitive '
        f'--db {db} -o {out} -p {args.threads} --outfmt 6 qseqid stitle evalue')
    subprocess.run(cmd, shell=True, stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL)

    datasetdb = str(Path(dfo, 'datasetdb/datasetdb.dmnd'))
    out2 = f'{args.output}/tmp/dataset_diamond.res'
    cmd2 = (
        f'diamond blastp -e 1e-10 -q {for_diamond} --more-sensitive '
        f'--db {datasetdb} -o {out2} -p {args.threads} --outfmt 6 qseqid stitle evalue')
    subprocess.run(cmd2, shell=True, stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL)


def correct_phylo_group(parent, sample_taxomomy):
    groups = set()
    for org in parent.get_leaf_names():
        if org in tax_group:
            groups.add(tax_group[org])
    if len(groups) > 2:
        return False
    elif sample_taxomomy in groups:
        return True
    return correct_phylo_group(parent.up, sample_taxomomy)


def fasttree(checked_hits):
    full_name = checked_hits[0].name
    org = full_name.split('_')[0]
    gene = full_name.split('@')[1]
    fas = f'{args.output}/tmp/{org}/{gene}.fas'
    aln = f'{args.output}/tmp/{org}/{gene}.aln'
    trim = f'{args.output}/tmp/{org}/{gene}.trimal'
    tree_file = f'{args.output}/tmp/{org}/{gene}.tree'
    copyfile(str(Path(dfo, f'orthologs/{gene}.fas')), f'{args.output}/tmp/{org}/{gene}.fas')
    with open(fas, 'a') as f:
        for hit in checked_hits:
            f.write(f'>{hit.name}\n{hit.seq}\n')
    cmd1 = f'mafft --auto --reorder {fas} > {aln}'
    subprocess.run(cmd1, shell=True, stderr=subprocess.DEVNULL)
    cmd2 = f"trimal -in {aln} -gt 0.2 -out {trim}"
    subprocess.run(cmd2, shell=True, stdout=subprocess.DEVNULL)
    cmd3 = f"fasttree {trim} > {tree_file}"
    subprocess.run(cmd3, shell=True, stderr=subprocess.DEVNULL)
    tree = Tree(tree_file)
    correct_len = length_check(trim)
    good_hits = []  # SBH hits
    good_hits_names = set()
    bb_hits = []  # all hits
    for hit in checked_hits:
        if hit.name in correct_len:
            bb_hits.append(hit)
            hit_node = tree.search_nodes(name=hit.name)[0]
            try:
                if correct_phylo_group(hit_node.up, input_taxonomy[org]) is True:
                    good_hits.append(hit)
                    good_hits_names.add(hit.name)
            except AttributeError:
                pass

    if args.all_bbh:
        # keep all hits, even if SBH (good hits) hits exist
        all_hits = good_hits[:]
        for hit in bb_hits:
            if hit.name not in good_hits_names:
                hit.name = hit.name.replace('_SBH', '_BBH')
                all_hits.append(hit)
        return all_hits
    elif len(good_hits) > 0:
        return good_hits
    elif len(bb_hits) > 0:
        for hit in bb_hits:
            hit.name = hit.name.replace('_SBH', '_BBH')
        return bb_hits
    else:
        return []


def parse_diamond_output():
    """checks if best hit from database is not bacterial or
    from wrong orthogroups

    returns: names of filtered hits with info about gene"""
    correct_hits = set()
    proccesed = set()
    for line in open(f'{args.output}/tmp/orthomcl_diamond.res'):
        sline = line.split("\t")
        full_name = sline[0]
        hit, gene = full_name.split('@')
        if full_name not in proccesed:
            proccesed.add(full_name)
            org = sline[1].split('|')[0]
            org = sline[1].split('|')[0]
            og = sline[1].split('|')[2].strip()
            if org not in bacterial and og in gene_og[gene]:
                correct_hits.add(full_name)
    return correct_hits


def new_best_hits(candidate_hits):
    """Final function. Prepares fasta files with result sequences.
    In case of _SBH path also checks phylogenetic posisiton of hits
    by fasttree function"""

    # if _SBH path: check for phylogenetically best candidates
    top_candidates = []
    if candidate_hits:
        if '_SBH' in candidate_hits[0].name and '*' not in input_taxonomy[candidate_hits[0].name.split('_')[0]]:
            top_candidates = fasttree(candidate_hits)
        else:
            top_candidates = candidate_hits

    # prepares final files with orthologs from dataset and new candidates
    if top_candidates:
        gene = top_candidates[0].name.split('@')[1]
        dataset = f'{args.output}/{gene}.fas'
        if not os.path.isfile(dataset):
            copyfile(str(Path(dfo, f'orthologs/{gene}.fas')), dataset)
        n = 0
        for cand in top_candidates:
            n += 1
            with open(dataset, 'a') as d:
                seq_name = cand.name.split("@")[0]
                gene = cand.name.split("@")[1]
                # TODO: is this exception handling just for debugging?
                try:
                    # if hit is reciprocal hit to a corresponding gene
                    if gene == reciprocal_hits[seq_name[:-4]]:
                        d.write(f'>{seq_name}_q{n}c\n{cand.seq}\n')
                    else:
                        d.write(f'>{seq_name}_q{n}n\n{cand.seq}\n')
                        with open(f'{args.output}/non_corresponding_hits.txt', 'a') as nonrep:
                            nonrep.write(
                                f'non-corresponding hit:{cand.name}; Best hit from:{reciprocal_hits[seq_name[:-4]]}\n')
                            print(f'non-corresponding hit:{cand.name}; Best hit from: {reciprocal_hits[seq_name[:-4]]}')
                except KeyError:
                    n -= 1


def parallel_new_best_hits(gene_hits):
    """Just pararallel run of new_best_candidates """
    with Pool(processes=int(args.threads)) as pool:
        pool.map(new_best_hits, gene_hits)


def prepare_good_hits():
    """Prepares all hits for filtration. It filters everything which is
    not in correct hits (wrongs orthogroups or bacterial best diamond
    hit from orthomcl database)"""
    gene_hits = defaultdict(list)
    for record in SeqIO.parse(f'{args.output}/tmp/for_diamond.fasta', 'fasta'):
        if record.name in correct_hits:
            gene = record.name.split('@')[1]
            org = record.name.split('_')[0]
            org_gene = f'{org}_{gene}'
            gene_hits[org_gene].append(record)
    parallel_new_best_hits(list(gene_hits.values()))


def get_reciprocal_hits():
    """Checks if candidate is reciprocal best blast hit with gene

    returns: list with reciprocal hits"""
    reciprocal = {}
    proccesed = set()
    for line in open(f'{args.output}/tmp/dataset_diamond.res'):
        sline = line.split("\t")
        full_name = sline[0]
        sequence, gene = full_name.split('@')  # do we need this line??
        sequence = sequence[:-4]
        if sequence not in proccesed:
            proccesed.add(sequence)
            hit_gene = sline[1]
            reciprocal[sequence] = hit_gene
    return reciprocal


def additions_to_input():
    """appends additional input to original input metadata"""
    original_input = config['PATHS']['input_file']
    with open(original_input, 'a') as ori:
        for line in open(args.add):
            if "File Name" not in line:
                ori.write(line)


if __name__ == '__main__':
    config = configparser.ConfigParser()
    config.read('config.ini')

    description = 'Script for ortholog fishing.'
    parser, optional, required = help_formatter.initialize_argparse(name='fisher.py',
                                                          desc=description,
                                                          usage='fisher.py [OPTIONS]')

    optional.add_argument('-t', '--threads', type=int, metavar='N', default=1,
                          help=textwrap.dedent("""\
                          Number of threads, where N is an integer.
                          Default: 1
                          """))
    optional.add_argument('-n', '--max_hits', type=int, metavar='N', default=5,
                          help=textwrap.dedent("""\
                          Max number of hits to check, where N is an integer.
                          Default: 5
                          """))
    optional.add_argument('--keep_tmp', action='store_true',
                          help=textwrap.dedent("""\
                          Keep temporary files
                          """))
    optional.add_argument('--all_bbh', action='store_true',
                          help=textwrap.dedent("""\
                          Keep all significant BLAST hits regardless of phylogenetic affiliation as 
                          determined with FastTree through the phylogenetically aware route.
                          """))
    optional.add_argument('--add', metavar='<inputfile>',
                          help=textwrap.dedent("""\
                          Input file (different from original one in config.ini) only with new organisms.
                          Requires --add_to option.
                          """))
    optional.add_argument('--add_to', metavar='<fisher_out>',
                          help=textwrap.dedent("""\
                          Previous output of fisher.py to add new organisms from a new input metadata.
                          Requires --add option.
                          """))
    args = help_formatter.get_args(parser, optional, required, pre_suf=False, inp_dir=False)

    if args.add and not args.add_to:
        parser.error("--add requires --add_to.")
    elif args.add_to and not args.add:
        parser.error("--add_to requires --add.")

    # database folder
    dfo = str(Path(config['PATHS']['database_folder']).resolve())

    # taxon: group for all orgs in metadata
    tax_group = taxonomy_dict()

    if args.add:
        # uses additional input metadata file
        input_metadata = os.path.abspath(args.add)
        args.output = os.path.abspath(args.add_to)
    else:
        # reads input metadata file from config
        input_metadata = os.path.abspath(config['PATHS']['input_file'])

    # checks input for potentional mistakes
    check_input()

    # reads bacterial names from orthocml database and also
    # information  abour gene: orthogroup for all genes
    bacterial, gene_og = bac_gog_db()

    # returns list with gene names
    profiles = get_hmm_profiles()

    if not args.add:
        # creates output dictionaries tmp and fasta
        makedirs()
    else:
        # if --add option
        if os.path.exists(f'{args.output}/tmp/'):
            try:
                # in a case that user used --keep-tmp option before:
                # it cleans problematic tmp files
                os.remove(f'{args.output}/tmp/dataset_diamond.res')
                os.remove(f'{args.output}/tmp/orthomcl_diamond.res')
                os.remove(f'{args.output}/tmp/for_diamond.fasta')
            except OSError:
                pass
        else:
            os.mkdir(f'{args.output}/tmp')

    # stores information about org:taxonomy for input samples
    input_taxonomy = {}

    for line in open(input_metadata):
        if "File Name" not in line:  # TODO fix me
            # input metadata file parsing
            metadata_input = line.split('\t')
            fasta_file = str(Path(metadata_input[0].strip(), metadata_input[1].strip()))
            sample_name = metadata_input[2].strip()
            taxonomy = metadata_input[3].strip()
            input_taxonomy[sample_name] = taxonomy
            specific_queries = metadata_input[5].strip()
            print(f'{sample_name} has started\n--------------------------')
            os.mkdir(f'{args.output}/tmp/{sample_name}')

            # performs cd-hit and rename sequences to common format
            # shortName_number
            infile = cluster_rename_sequences()

            # parses specifiq queries
            specific_queries = specific_queries.strip()
            if specific_queries.lower() == 'none':
                specific_queries = None

            # prepares candidates for all genes (according to blast and/or hmmsearch)
            # and stores them in for_diamond.fasta
            phylofisher(args.threads,
                        args.max_hits, specific_queries)

    # performs diamond search against orthomcl database
    diamond()
    # parses diamond output and returns filtered hits
    # with information about gene in the header
    # (no bacterial, or wrong orthogroups)
    correct_hits = parse_diamond_output()
    # returns hits which are reciprocal with gene from datasetdb
    reciprocal_hits = get_reciprocal_hits()
    # final function
    prepare_good_hits()

    if args.add:
        # adds additional input metadata to original input metadata
        additions_to_input()
    # when user doesn't want to keep tmp/* files
    if not args.keep_tmp:
        rmtree(f'{args.output}/tmp/')
    print('Fisher completed without any errors.')



#!/usr/bin/env python
import os
import shutil
import subprocess

from phylofisher import help_formatter


def make_astral_inputs():
    """

    :return:
    """
    genes = [file.split('.')[1] for file in os.listdir(args.input) if 'RAxML_bipartitions.' in file]
    genes = set(genes)
    
    with open(f'{args.output}/all_sgt.tre', 'w') as tree_file, open(f'{args.output}/bs_files.txt', 'w') as bs_files:
        for gene in genes:
            with open(f'{args.input}/RAxML_bipartitions.{gene}.{args.suffix}', "r") as infile:
                for line in infile:
                    line = line.strip()
                    tree_file.write(f'{line}\n')
            
            bs_files.write(f'{args.input}/RAxML_bootstrap.{gene}.{args.suffix}\n')
                    

def run_astral():
    """

    :return:
    """
    os.chdir(args.output)
    subprocess.run(f'astral -i all_sgt.tre -b bs_files.txt -o AstralBS.out', shell=True, executable='/bin/bash')


if __name__ == '__main__':
    description = 'Generates input files and infers a coalescent-based species tree'  # TODO: Get Description
    parser, optional, required = help_formatter.initialize_argparse(name='astral_runner.py',
                                                                    desc=description,
                                                                    usage='astral_runner.py '
                                                                          '[OPTIONS] -i /path/to/input/')

    # Optional Arguments
    # optional.add_argument('--in_format', metavar='<format>', type=str, default='fasta',
    #                       help=textwrap.dedent("""\
    #                               Desired format of the output steps.
    #                               Options: fasta, nexus, phylip (names truncated at 10 characters),
    #                               or phylip-relaxed (names are not truncated)
    #                               Default: phylip-relaxed
    #                               """))

    in_help = 'Path to directory containing single gene trees and their corresponding bootstrap value files.'
    args = help_formatter.get_args(parser, optional, required, in_help=in_help)
    
    args.output = os.path.abspath(args.output)
    args.input = os.path.abspath(args.input)
    
    if os.path.isdir(args.output):
        shutil.rmtree(args.output)
    os.mkdir(args.output)

    make_astral_inputs()
    run_astral()



#!/usr/bin/env python
import csv
import os
import shutil
import subprocess
import sys
import textwrap
from collections import defaultdict
from glob import glob
from multiprocessing import Pool

from Bio import SeqIO
from Bio.Seq import Seq
from Bio.SeqRecord import SeqRecord

from phylofisher import help_formatter


def bash(cmd):
    subprocess.run(cmd, executable='/bin/bash', shell=True)


def mk_dirs():
    """

    :return:
    """
    dirs = ['prequal', 'mafft', 'divvier', 'trimal']

    for my_dir in dirs:
        my_dir = f'{args.output}/{my_dir}'
        if os.path.isdir(my_dir) is False:
            os.mkdir(my_dir)


def delete_gaps_stars(gene, root):
    """Removes -'s and *'s from alignments"""
    file_name = f'{root}.aa'
    records = []
    for record in SeqIO.parse(gene, args.in_format):
        seq_string = str(record.seq).replace("-", "").replace("*", "")
        records.append(SeqRecord(Seq(seq_string),
                                 id='',
                                 name='',
                                 description=record.description))

    with open(file_name, 'w') as res:
        SeqIO.write(records, res, 'fasta')


def trim_and_align(gene):
    root = os.path.basename(gene).split('.')[0]
    # prequal
    os.chdir(f'{args.output}/prequal')
    delete_gaps_stars(gene, root)
    bash(f'prequal {root}.aa')
    os.chdir(args.output)

    # mafft
    os.chdir(f'{args.output}/mafft')
    bash(f'mafft --globalpair --maxiterate 1000 --unalignlevel 0.6 '
         f'--thread 1 {args.output}/prequal/{root}.aa.filtered > {root}.aln')
    os.chdir(args.output)

    # divvier
    os.chdir(f'{args.output}/divvier')
    bash(f'divvier -partial -mincol 4 -divvygap {args.output}/mafft/{root}.aln')
    shutil.move(f'{args.output}/mafft/{root}.aln.partial.fas', f'{args.output}/divvier/{root}.aln.partial.fas')
    shutil.move(f'{args.output}/mafft/{root}.aln.PP', f'{args.output}/divvier/{root}.aln.PP')
    os.chdir(args.output)

    # trimal
    os.chdir(f'{args.output}/trimal')
    bash(f'trimal -in {args.output}/divvier/{root}.aln.partial.fas -gt 0.80 -out {root}.gt80trimal.fas -fasta')
    os.chdir(args.output)


def parallelize(files):
    processes = args.threads
    if len(files) < args.threads:
        processes = len(files)

    with Pool(processes=processes) as p:
        all_checks = p.map(trim_and_align, files)


def parse_names(input_folder):
    """

    :param input_folder:
    :return:
    """
    name_set = set()
    if args.suffix:
        files = sorted(glob(f'{input_folder}/*{args.suffix}'))
    else:
        files = sorted(glob(f'{input_folder}/*'))
    for file in files:
        with open(file) as f:
            for record in SeqIO.parse(f, args.in_format):
                fname = record.description
                name = fname.split('_')[0]
                name_set.add(name)
    return files, sorted(list(name_set))


def stats(total_len, out_dict):
    """
    tools.py [OPTIONS] -i <input_dir> -m <metadata> {-n <gene_number> | -c <percent_complete>}

    :param total_len:
    :return:
    """
    with open(f'{args.output}/matrix_constructor_stats.tsv', 'w') as out_file:
        tsv_writer = csv.writer(out_file, delimiter='\t')
        tsv_writer.writerow(['Taxon', 'PercentMissingData'])
        missing = []
        for record in SeqIO.parse(f'{args.output}/matrix.{out_dict[args.out_format.lower()]}', args.out_format.lower()):
            missing.append((record.name, (record.seq.count('-') / total_len) * 100))
        for org_missing in sorted(missing, key=lambda x: x[1], reverse=True):
            tsv_writer.writerow(list(org_missing))


if __name__ == '__main__':
    description = 'To trim align and concatenate orthologs into a super-matrix run:'
    parser, optional, required = help_formatter.initialize_argparse(name='matrix_constructor.py',
                                                                    desc=description,
                                                                    usage='matrix_constructor.py [OPTIONS] -i path/to/input/')

    # Optional Arguments
    optional.add_argument('-of', '--out_format', metavar='<format>', type=str, default='fasta',
                          help=textwrap.dedent("""\
                          Desired format of the output matrix.
                          Options: fasta, phylip (names truncated at 10 characters), 
                          phylip-relaxed (names are not truncated), or nexus.
                          Default: fasta
                          """))
    optional.add_argument('-if', '--in_format', metavar='<format>', type=str, default='fasta',
                          help=textwrap.dedent("""\
                          Format of the input files.
                          Options: fasta, phylip (names truncated at 10 characters), 
                          phylip-relaxed (names are not truncated), or nexus.
                          Default: fasta
                          """))
    optional.add_argument('-c', '--concatenation_only', action='store_true', default=False,
                          help=textwrap.dedent("""\
                          Only concatenate alignments. Trimming and alignment are not performed automatically.
                          """))
    optional.add_argument('-t', '--threads', metavar='N', type=int, default=1,
                          help=textwrap.dedent("""\
                          Desired number of threads to be utilized.
                          Default: 1
                          """))

    # Changes help descriptions from the default input and output help descriptions
    in_help = 'Path to prep_final_dataset_<M.D.Y>'
    args = help_formatter.get_args(parser, optional, required, in_help=in_help)

    args.input = os.path.abspath(args.input)
    args.output = os.path.abspath(args.output)

    try:
        os.mkdir(args.output)
    except OSError:
        shutil.rmtree(args.output)
        os.mkdir(args.output)

    files, orgs = parse_names(args.input)
    if args.concatenation_only is False:
        os.chdir(args.output)
        mk_dirs()
        parallelize(files)
        files = sorted(glob(f'{args.output}/trimal/*'))

    total_len = 0
    res_dict = defaultdict(str)
    with open(f'{args.output}/indices.tsv', 'w') as outfile:
        outfile.write('Gene\tStart\tStop\n'
                      '')
        for file in files:
            gene = os.path.basename(file).split('.')[0]
            length = 0
            seq_dict = {}
            if args.concatenation_only:
                myformat = args.in_format
            else:
                myformat = 'fasta'
            for record in SeqIO.parse(file, myformat):
                length = len(record.seq)
                seq_dict[record.id.split('_')[0]] = str(record.seq)
            start_len = total_len + 1
            total_len += length
            outfile.write(f'{gene}\t{start_len}\t{total_len}\n')
            for org in orgs:
                if org in seq_dict:
                    res_dict[org] += seq_dict[org]
                else:
                    res_dict[org] += ('-' * length)

    # Accepted out formats with respective suffix
    out_dict = {'fasta':          'fas',
                'phylip':         'phy',
                'phylip-relaxed': 'phy',
                'nexus':          'nex'}

    # Creates SeqRecord iterator
    records = []
    for org, seq in res_dict.items():
        records.append(SeqRecord(Seq(seq),
                                 id=org,
                                 name='',
                                 description=''))

    # Writes to output matrix in user specified output
    if args.out_format.lower() in out_dict:
        with open(f'{args.output}/matrix.{out_dict[args.out_format.lower()]}', "w") as handle:
            SeqIO.write(records, handle, args.out_format.lower())
    else:
        sys.exit('Invalid Output Format')

    stats(total_len, out_dict)
